#/*
# * Copyright (c) 2020 Xilinx Inc. All rights reserved.
# *
# * Author:
# *       Appana Durga Kedareswara rao <appana.durga.rao@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */
import sys
import os
import re
import yaml

sys.path.append(os.path.dirname(__file__))

import common_utils as utils
from domain_access import domain_get_subnodes

def get_label(sdt, symbol_node, node):
    prop_dict = symbol_node.__props__
    match = [label for label,node_abs in prop_dict.items() if re.match(node_abs[0], node.abs_path) and len(node_abs[0]) == len(node.abs_path)]
    if match:
        return match[0]
    else:
        return None

def get_cpu_node(sdt, options):
    cpu_name = options['args'][0]
    symbol_node = sdt.tree['/__symbols__']
    nodes = sdt.tree.nodes('/cpu.*')
    cpu_labels = []
    matched_label = None
    for node in nodes:
        matched_label = get_label(sdt, symbol_node, node)
        if matched_label is not None:
            if matched_label == cpu_name:
                return node
            elif node.propval('reg') != ['']:
                cpu_labels.append(matched_label)

    print(f"ERROR: In valid CPU Name valid Processors for a given SDT are {cpu_labels}\n")
    sys.exit(1)

def item_generator(json_input, lookup_key):
    if isinstance(json_input, dict):
        for k, v in json_input.items():
            if k == lookup_key:
                if isinstance(v, str):
                    yield [v]
                else:
                    yield v
            else:
                for child_val in item_generator(v, lookup_key):
                    yield child_val
    elif isinstance(json_input, list):
        for item in json_input:
            for item_val in item_generator(item, lookup_key):
                yield item_val

# This API reads the schema and returns the compatible list
def compat_list(schema):
    compatible_list = []
    if schema and 'compatible' in schema.get('properties',{}).keys():
        sch = schema['properties']['compatible']
        compatible_list = []
        for l in item_generator(sch, 'enum'):
            compatible_list.extend(l)

        for l in item_generator(sch, 'const'):
            compatible_list.extend(l)

        if 'contains' in sch.keys():
            for l in item_generator(sch['contains'], 'enum'):
                compatible_list.extend(l)

            for l in item_generator(sch['contains'], 'const'):
                compatible_list.extend(l)
        compatible_list = list(set(compatible_list))
    return compatible_list

"""
This API scans the device-tree node and returns the address
and size of the reg property for the user provided index.

Args:
    node: LopperNode object
    value: property value
    idx: index
"""
def scan_reg_size(node, value, idx):
    na = node.parent["#address-cells"].value[0]
    ns = node.parent["#size-cells"].value[0]
    cells = na + ns
    reg = 0
    size = 0
    if cells > 2:
        reg1 = value[cells * idx]
        if reg1 != 0:
            val = str(hex(value[cells * idx + 1]))[2:]
            pad = 8 - len(val)
            val = val.ljust(pad + len(val), '0')
            reg = int((str(hex(reg1)) + val), base=16)
        else:
            reg = value[cells * idx + 1]

        size1 = value[cells * idx + na]
        if size1 != 0:
            val = str(hex(value[cells * idx + ns + 1]))[2:]
            pad = 8 - len(val)
            val = val.ljust(pad + len(val), '0')
            size = int((str(hex(size1)) + val), base=16)
        else:
            size = value[cells * idx + ns + 1]
    elif cells == 2:
        reg = value[idx * cells]
        size = value[idx * cells + 1]
    else:
        reg = value[0]
    return reg, size

def get_interrupt_prop(sdt, node, value):
    intr = []
    inp =  node['interrupt-parent'].value[0]
    intr_parent = [node for node in sdt.tree['/'].subnodes() if node.phandle == inp]
    inc = intr_parent[0]["#interrupt-cells"].value[0]
    """
    Baremetal Interrupt Property format:
        bits[11:0]  interrupt-id
        bits[15:12] trigger type and level flags
        bits[19:16] CPU Mask
        bit[20] interrupt-type (1: PPI, 0: SPI)
    """
    # Below logic converts the interrupt propery value to baremetal
    # interrupt property format.
    nintr = len(value)/inc
    tmp = inc % 2
    for val in range(0, int(nintr)):
        intr_sensitivity = value[tmp+1] << 12
        intr_id = value[tmp]
        # Convert PPI interrupt to baremetal interrupt format
        if value[tmp-1] == 1 and inc == 3:
            intr_sensitivity = (value[tmp+1] & 0xF) << 12
            cpu_mask = (value[tmp+1] & 0xFF00) << 8
            ppi_type = 1 << 20
            intr_id = intr_id + cpu_mask + ppi_type
        intr.append(hex(intr_id + intr_sensitivity))
        tmp += inc

    return intr

#Return the base address of the parent node.
def get_phandle_regprop(sdt, prop, value):
    parent_node = [node for node in sdt.tree['/'].subnodes() if node.phandle == value[0]]
    reg, size = scan_reg_size(parent_node[0], parent_node[0]['reg'].value, 0)
    # Special handling for Soft Ethernet(1/2.5G, and 10G/25G MAC) axistream-connected property
    if prop == "axistream-connected":
        compat = parent_node[0]['compatible'].value
        axi_fifo = [item for item in compat if "xlnx,axi-fifo" in item]
        axi_dma = [item for item in compat if "xlnx,eth-dma" in item]
        axi_mcdma = [item for item in compat if "xlnx,eth-mcdma" in item]
        if axi_fifo:
            reg += 1
        elif axi_dma:
            reg += 2
        elif axi_mcdma:
            reg += 3
    return reg

#Return the base address of the interrupt parent.
def get_intrerrupt_parent(sdt, value):
    intr_node = [node for node in sdt.tree['/'].subnodes() if node.phandle == value[0]]
    reg, size = scan_reg_size(intr_node[0], intr_node[0]['reg'].value, 0)
    """
    Baremetal Interrupt Parent Property Format:
        bits[0]    Interrupt parent type (0: GIC, 1: AXI INTC)
        bits[31:1] Base Address of the interrupt parent
    """
    compat = intr_node[0]['compatible'].value
    axi_intc = [item for item in compat if "xlnx,xps-intc-1.00.a" in item]
    if axi_intc:
        reg += 1
    return reg

"""
This API scans the device-tree node and returns the address
and size of the ranges property for the user provided index.

Args:
    node: LopperNode object
    value: property value
    idx: index
"""
def scan_ranges_size(node, value, idx):
    na = node["#address-cells"].value[0]
    ns = node["#size-cells"].value[0]
    cells = na + ns + 2

    addr = 0
    size = 0

    addr1 = value[cells * idx + 1]
    if addr1 != 0:
        val = str(value[cells * idx + ns])
        pad = 8 - len(val)
        val = val.ljust(pad + len(val), '0')
        addr = int((str(hex(addr1)) + val), base=16)
    else:
        addr = value[cells * idx + ns]

    size1 = value[cells * idx + na + 2]
    if size1 != 0:
        val = str(hex(value[cells * idx + na + 3]))[2:]
        pad = 8 - len(str(hex(size1)))
        val = val.ljust(pad + len(str(hex(size1))), '0')
        size = int((str(hex(size1)) + val), base=16)
    else:
        size = value[cells * idx + na + ns + 1]
    return addr, size

def get_clock_prop(sdt, value):
    clk_node = [node for node in sdt.tree['/'].subnodes() if node.phandle == value[0]]
    """
    Baremetal clock format:
        bits[0] clock parent(controller) type(0: ZynqMP clock controller)
        bits[31:1] clock value
    """
    compat = clk_node[0]['compatible'].value
    return value[1]

def get_pci_ranges(node, value, pad):
    pci_ranges = []
    for i in range(pad):
        try:
            reg, size = scan_ranges_size(node, value, i)
            high_addr = reg + size - 1
            pci_ranges.append(hex(reg))
            pci_ranges.append(hex(high_addr))
        except IndexError:
            pci_ranges.append(hex(0))
            pci_ranges.append(hex(0))
    return pci_ranges

class DtbtoCStruct(object):
    def __init__(self, out_file):
        self._outfile = open(out_file, 'w')
        self._lines = []

    def out(self, line):
        """Output a string to the output file

        Args:
            line: String to output
        """
        self._outfile.write(line)

    def buf(self, line):
        """Buffer up a string to send later

        Args:
            line: String to add to our 'buffer' list
        """
        self._lines.append(line)

    def get_buf(self):
        """Get the contents of the output buffer, and clear it

        Returns:
            The output buffer, which is then cleared for future use
        """
        lines = self._lines
        self._lines = []
        return lines

def is_compat(node, compat_string_to_test):
    if re.search( "module,baremetalconfig_xlnx", compat_string_to_test):
        return xlnx_generate_bm_config
    return ""

def get_stdin(sdt, chosen_node, node_list):
    prop_val = chosen_node['stdout-path'].value
    serial_node = sdt.tree.alias_node(prop_val[0].split(':')[0])
    match = [x for x in node_list if re.search(x.name, serial_node.name)]
    return match[0]

def get_mapped_nodes(sdt, node_list, options):
    # Yocto Machine to CPU compat mapping
    match_cpunode = get_cpu_node(sdt, options)

    all_phandles = []
    address_map = match_cpunode.parent["address-map"].value
    na = match_cpunode.parent["#ranges-address-cells"].value[0]
    ns = match_cpunode.parent["#ranges-size-cells"].value[0]
    cells = na + ns
    tmp = na
    while tmp < len(address_map):
        all_phandles.append(address_map[tmp])
        tmp = tmp + cells + na + 1

    # Get all Domains and see if any other peripheral is accessing the mapped node
    # If Mapped delete it from valid_nodes
    domain_nodes = domain_get_subnodes(sdt.tree)

    # Get all nodes in that remove BM domain nodes
    invalid_phandles = []
    shared_phandles = []
    for domain_node in domain_nodes:
        bm_domain = []
        if domain_node.propval('os,type') != ['']:
            if re.search('baremetal', domain_node.propval('os,type', list)[0]):
                    bm_domain.append(domain_node)
        elif domain_node.propval('cpus') != ['']:
                print('ERROR: os,type property is missing in the domain', domain_node.name)

        if bm_domain:
            if domain_node['cpus'].value[0] == match_cpunode.parent.phandle:
               # Baremetal access property
               # Check for shared resources
               if domain_node.propval('access') != ['']:
                   shared_phandles.extend(domain_node['access'].value)

               if domain_node.propval('include') != ['']:
                   rsc_domain = domain_node['include'].value
                   for rsc in rsc_domain:
                       match = [node for node in domain_nodes if node.phandle == rsc]
                       if match:
                           try:
                               shared_phandles.extend(match[0]['access'].value)
                           except:
                               pass
            else:
                if domain_node.propval('access') != ['']:
                   invalid_phandles.extend(domain_node['access'].value)
        else:
            if domain_node.propval('access') != ['']:
                invalid_phandles.extend(domain_node['access'].value)

    # Remove duplicate phandle
    all_phandles = list(dict.fromkeys(all_phandles))
    invalid_phandles = list(dict.fromkeys(invalid_phandles))

    """
    Create a valid node list, Here valid node list means
    1) Without any domains cpu cluster address-map property mapped nodes are valid nodes
    2) With domains below are assumptions
        i) If the domain node has access property then it takes priority than address-map property
           (i.e if any other domain node contains access property and the same peripheral node is
            not mapped in our domain access node then it should be removed from the valid_node list)
        ii) If the same peripheral node is mapped in our domain and other domain access node then
            the node should include it in the valid_node list,
        iii) If the peripheral node is in a shared resource and if it is mapped for the baremetal domain
             then include it in the valid node list else remove it from the valid node list.
    """
    # Remove shared phandles from invalid phandles list
    invalid_phandles_list = [phandle for phandle in invalid_phandles if phandle not in shared_phandles]
    valid_phandles = [phandle for phandle in all_phandles if phandle not in invalid_phandles_list]
    valid_nodes = [node for node in node_list for handle in valid_phandles if handle == node.phandle]
    return valid_nodes

# tgt_node: is the baremetal config top level domain node number
# sdt: is the system device-tree
# options: baremetal driver meta-data file path
def xlnx_generate_bm_config(tgt_node, sdt, options):
    root_node = sdt.tree[tgt_node]
    root_sub_nodes = root_node.subnodes()
    node_list = []
    chosen_node = ""
    # Traverse the tree and find the nodes having status=ok property
    for node in root_sub_nodes:
        try:
            if node.name == "chosen":
                chosen_node = node
            status = node["status"].value
            if "okay" in status:
                node_list.append(node)
        except:
           pass

    src_dir = options['args'][1]
    stdin = ""
    try:
        stdin = options['args'][2]
        stdin_node = get_stdin(sdt, chosen_node, node_list)
    except IndexError:
        pass

    drvpath = utils.get_dir_path(src_dir.rstrip(os.sep))
    drvname = utils.get_base_name(drvpath)
    yaml_file = os.path.join(drvpath, f"data/{drvname}.yaml")
    if not utils.is_file(yaml_file):
        print(f"{drvname} Driver doesn't have yaml file")
        return False

    driver_compatlist = []
    # Read the yaml file and get the driver supported compatible list
    # and config data file required driver properties
    schema = utils.load_yaml(yaml_file)
    driver_compatlist = compat_list(schema)
    driver_proplist = schema.get('required',[])
    config_struct = schema.get('config',[])
    driver_optproplist = schema.get('optional',[])

    driver_nodes = []
    for compat in driver_compatlist:
        for node in node_list:
           compat_string = node['compatible'].value
           for compa in compat_string:
               if compat in compa:
                   driver_nodes.append(node)

    # Remove duplicate nodes
    driver_nodes = list(set(driver_nodes))
    driver_nodes = get_mapped_nodes(sdt, driver_nodes, options)
    if not config_struct:
        config_struct = str("X") + drvname.capitalize() + str("_Config")
    else:
        config_struct = config_struct[0]
        drvname = config_struct.split('_Config')[0].lower()
        drvname = drvname[1:]
    outfile = os.path.join(sdt.outdir,f"x{drvname}_g.c")

    plat = DtbtoCStruct(outfile)
    nodename_list = []
    for node in driver_nodes:
        nodename_list.append(node.name)

    cmake_file = drvname.upper() + str("Config.cmake")
    cmake_file = os.path.join(sdt.outdir,f"{cmake_file}")
    with open(cmake_file, 'a') as fd:
       fd.write("set(DRIVER_INSTANCES %s)\n" % utils.to_cmakelist(nodename_list))
       if stdin:
           match = [x for x in nodename_list if re.search(x, stdin_node.name)]
           if match:
               fd.write("set(STDIN_INSTANCE %s)\n" % '"{}"'.format(match[0]))

    for index,node in enumerate(driver_nodes):
        drvprop_list = []
        drvoptprop_list = []
        if index == 0:
            plat.buf('#include "x%s.h"\n' % drvname)
            plat.buf('\n%s %s __attribute__ ((section (".drvcfg_sec"))) = {\n' % (config_struct, config_struct + str("Table[]")))
        for i, prop in enumerate(driver_proplist):
            pad = 0
            phandle_prop = 0
            # Few drivers has multiple data interface type (AXI4 or AXI4-lite),
            # Driver config structures of these SoftIP's contains baseaddress entry for each possible data interface type.
            # Device-tree node reg property may or may not contain all the possible entries that driver config structure
            # is expecting, In that case we need to add dummy entries(0xFF) in the config structure in order to avoid
            # compilation errors.
            #
            # Yaml meta-data representation/syntax will be like below
            # reg: <range of baseaddress>
            # interrupts: <supported range of interrupts>
            if isinstance(prop, dict):
               pad = list(prop.values())[0]
               prop = list(prop.keys())[0]
               if pad == "phandle":
                   phandle_prop = 1
            if i == 0:
                 plat.buf('\n\t{')

            if prop == "reg":
                val, size = scan_reg_size(node, node[prop].value, 0)
                drvprop_list.append(hex(val))
                plat.buf('\n\t\t%s' % hex(val))
                if pad:
                    for j in range(1, pad):
                        try:
                            val, size = scan_reg_size(node, node[prop].value, j)
                            drvprop_list.append(hex(val))
                            plat.buf(',\n\t\t%s' % hex(val))
                        except IndexError:
                            plat.buf(',\n\t\t%s' % hex(0xFFFF))
            elif prop == "compatible":
                plat.buf('\n\t\t%s' % '"{}"'.format(node[prop].value[0]))
                drvprop_list.append(node[prop].value[0])
            elif prop == "interrupts":
                try:
                    intr = get_interrupt_prop(sdt, node, node[prop].value)
                except KeyError:
                    intr = [hex(0xFFFF)]

                if pad:
                    plat.buf('\n\t\t{')
                    for j in range(0, pad):
                        try:
                            plat.buf('%s' % intr[j])
                            drvprop_list.append(intr[j])
                        except IndexError:
                            plat.buf('%s' % hex(0xFFFF))
                            drvprop_list.append(hex(0xFFFF))
                        if j != pad-1:
                            plat.buf(',  ')
                    plat.buf('}')
                else:
                    plat.buf('\n\t\t%s' % intr[0])
                    drvprop_list.append(intr[0])
            elif prop == "interrupt-parent":
                try:
                    intr_parent = get_intrerrupt_parent(sdt, node[prop].value)
                except KeyError:
                    intr_parent = 0xFFFF
                plat.buf('\n\t\t%s' % hex(intr_parent))
                drvprop_list.append(hex(intr_parent))
            elif prop == "clocks":
                clkprop_val = get_clock_prop(sdt, node[prop].value)
                plat.buf('\n\t\t%s' % hex(clkprop_val))
                drvprop_list.append(hex(clkprop_val))
            elif prop == "child,required":
                plat.buf('\n\t\t{')
                for j,child in enumerate(list(node.child_nodes.items())):
                    if len(pad) != 1:
                        plat.buf('\n\t\t\t{')
                    for k,p in enumerate(pad):
                        try:
                            plat.buf('\n\t\t\t\t%s' % hex(child[1][p].value[0]))
                            drvprop_list.append(hex(child[1][p].value[0]))
                        except KeyError:
                            plat.buf('\n\t\t\t\t%s' % hex(0xFFFF))
                            drvprop_list.append(hex(0xFFFF))
                        if k != (len(pad) - 1) or len(pad) == 1:
                            plat.buf(',')
                        plat.buf(' /* %s */' % p)
                    if len(pad) != 1:
                        if j != (len(list(node.child_nodes.items())) - 1):
                            plat.buf('\n\t\t\t},')
                        else:
                            plat.buf('\n\t\t\t}')
                plat.buf('\n\t\t}')
            elif phandle_prop:
                try:
                    prop_val = get_phandle_regprop(sdt, prop, node[prop].value)
                except KeyError:
                    prop_val = 0
                plat.buf('\n\t\t%s' % hex(prop_val))
                drvprop_list.append(hex(prop_val))
            elif prop == "ranges":
                try:
                    device_type = node['device_type'].value[0]
                    if device_type == "pci":
                        device_ispci = 1
                except KeyError:
                    device_ispci = 0
                if device_ispci:
                    prop_vallist = get_pci_ranges(node, node[prop].value, pad)
                    for j, prop_val in enumerate(prop_vallist):
                        plat.buf('\n\t\t%s' % prop_val)
                        if j != (len(prop_vallist) - 1):
                            plat.buf(',')
                        drvprop_list.append(prop_val)
            else:
                try:
                    prop_val = node[prop].value
                    # For boolean property if present LopperProp will return
                    # empty string convert it to baremetal config struct expected value
                    if '' in prop_val:
                        prop_val = [1]
                except KeyError:
                    prop_val = [0]

                if ('/bits/' in prop_val):
                    prop_val = [int(prop_val[-1][3:-1], base=16)]

                if len(prop_val) > 1:
                    plat.buf('\n\t\t{')
                    for k,item in enumerate(prop_val):
                        if isinstance(item, int):
                            drvprop_list.append(hex(item))
                        else:
                            drvprop_list.append(item)
                        plat.buf('%s' % item)
                        if k != len(prop_val)-1:
                            plat.buf(',  ')
                    plat.buf('}')
                else:
                    drvprop_list.append(hex(prop_val[0]))
                    plat.buf('\n\t\t%s' % hex(prop_val[0]))

            if i == len(driver_proplist)-1:
                plat.buf(' /* %s */' % prop)
                plat.buf('\n\t},')
            else:
                plat.buf(',')
                plat.buf(' /* %s */' % prop)
        if index == len(driver_nodes)-1:
            plat.buf('\n\t {\n\t\t NULL\n\t}')
            plat.buf('\n};')

        for i, prop in enumerate(driver_optproplist):
            if isinstance(prop, dict):
               pad = list(prop.values())[0]
               prop = list(prop.keys())[0]
            if prop == "child,required":
                for j,child in enumerate(list(node.child_nodes.items())):
                    for k,p in enumerate(pad):
                        drvoptprop_list.append(child[1][p].value[0])
            else:
                try:
                    drvoptprop_list.append(hex(node[prop].value[0]))
                except KeyError:
                    pass

        with open(cmake_file, 'a') as fd:
           fd.write("set(DRIVER_PROP_%s_LIST %s)\n" % (index, utils.to_cmakelist(drvprop_list)))
           fd.write("set(DRIVER_OPTPROP_%s_LIST %s)\n" % (index, utils.to_cmakelist(drvoptprop_list)))
           fd.write("list(APPEND TOTAL_DRIVER_PROP_LIST DRIVER_PROP_%s_LIST)\n" % index)
    plat.out(''.join(plat.get_buf()))

    return True



#/*
# * Copyright (c) 2020 Xilinx Inc. All rights reserved.
# *
# * Author:
# *       Appana Durga Kedareswara rao <appana.durga.rao@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import sys
import os
import re

sys.path.append(os.path.dirname(__file__))

import common_utils as utils
import baremetalconfig_xlnx as bm_config
from baremetaldrvlist_xlnx import xlnx_generate_bm_drvlist
from baremetallinker_xlnx import get_memranges

def is_compat( node, compat_string_to_test ):
    if re.search( "module,baremetal_xparameters_xlnx", compat_string_to_test):
        return xlnx_generate_xparams
    return ""

def xlnx_generate_xparams(tgt_node, sdt, options):
    root_node = sdt.tree[tgt_node]
    root_sub_nodes = root_node.subnodes()

    node_dict = {}
    node_list = []
    # Traverse the tree and find the nodes having status=ok property
    symbol_node = ""
    chosen_node = ""
    for node in root_sub_nodes:
        try:
            if node.name == "chosen":
                chosen_node = node
            if node.name == "__symbols__":
                symbol_node = node
            status = node["status"].value
            if "okay" in status:
                node_list.append(node)
                compat = node['compatible'].value
                node_dict.update({node.abs_path:compat})
        except:
           pass

    repo_path_data = options['args'][1]
    for node in node_list:
        try:
            prop_val = node['dma-coherent'].value
            if '' in prop_val:
                cci_en = 1
                break
        except KeyError:
            cci_en = None

    drvlist = xlnx_generate_bm_drvlist(tgt_node, sdt, options)
    xparams = os.path.join(sdt.outdir, f"xparameters.h")
    plat = bm_config.DtbtoCStruct(xparams)
    plat.buf('#ifndef XPARAMETERS_H   /* prevent circular inclusions */\n')
    plat.buf('#define XPARAMETERS_H   /* by using protection macros */\n')
    total_nodes = node_list
    for drv in drvlist:
        if utils.is_file(repo_path_data):
            repo_schema = utils.load_yaml(repo_path_data)
            drv_data = repo_schema['driver']
            drv_dir = drv_data.get(drv,{}).get('vless','')
        else:
            drv_dir = os.path.join(repo_path_data, "XilinxProcessorIPLib", "drivers", drv)

        yaml_file_abs = os.path.join(drv_dir, "data", f"{drv}.yaml")

        if utils.is_file(yaml_file_abs):
            schema = utils.load_yaml(yaml_file_abs)
            driver_compatlist = bm_config.compat_list(schema)
            driver_proplist = schema.get('required',{})
            match_nodes = []
            for comp in driver_compatlist:
                for node,compatible_list in sorted(node_dict.items(), key=lambda e: e[0], reverse=False):
                   match = [x for x in compatible_list if comp == x]
                   if match:
                       node1 = [x for x in node_list if (x.abs_path == node)]
                       node_list = [x for x in node_list if not(x.abs_path == node)]
                       if node1:
                           match_nodes.append(node1[0])
            match_nodes = bm_config.get_mapped_nodes(sdt, match_nodes, options)
            for index, node in enumerate(match_nodes):
                label_name = bm_config.get_label(sdt, symbol_node, node)
                label_name = label_name.upper()
                canondef_dict = {}
                if index == 0:
                    plat.buf(f'\n#define XPAR_X{drv.upper()}_NUM_INSTANCES {len(match_nodes)}\n')
                for i, prop in enumerate(driver_proplist):
                    pad = 0
                    phandle_prop = 0
                    if isinstance(prop, dict):
                        pad = list(prop.values())[0]
                        prop = list(prop.keys())[0]
                        if pad == "phandle":
                            phandle_prop = 1

                    if i == 0:
                        plat.buf(f'\n/* Definitions for peripheral {label_name} */')

                    if prop == "reg":
                        try:
                            val, size = bm_config.scan_reg_size(node, node[prop].value, 0)
                            plat.buf(f'\n#define XPAR_{label_name}_BASEADDR {hex(val)}')
                            plat.buf(f'\n#define XPAR_{label_name}_HIGHADDR {hex(val + size -1)}')
                            canondef_dict.update({"BASEADDR":hex(val)})
                            canondef_dict.update({"HIGHADDR":hex(val + size - 1)})
                            if pad:
                                for j in range(1, pad):
                                    try:
                                        val, size = bm_config.scan_reg_size(node, node[prop].value, j)
                                        plat.buf(f'\n#define XPAR_{label_name}_BASEADDR_{j} {hex(val)}')
                                    except IndexError:
                                        pass
                        except KeyError:
                            pass
                    elif prop == "compatible":
                        plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()} {node[prop].value[0]}')
                        canondef_dict.update({prop:node[prop].value[0]})
                    elif prop == "interrupts":
                        try:
                            intr = bm_config.get_interrupt_prop(sdt, node, node[prop].value)
                            plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()} {intr[0]}')
                            canondef_dict.update({prop:intr[0]})
                        except KeyError:
                            intr = [0xFFFF]

                        if pad:
                            for j in range(1, pad):
                                try:
                                    plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()}_{j} {intr[j]}')
                                except IndexError:
                                    pass
                    elif prop == "interrupt-parent":
                        try:
                            intr_parent = bm_config.get_intrerrupt_parent(sdt, node[prop].value)
                            prop = prop.replace("-", "_")
                            plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()} {hex(intr_parent)}')
                            canondef_dict.update({prop:hex(intr_parent)})
                        except KeyError:
                            pass
                    elif prop == "clocks":
                        clkprop_val = bm_config.get_clock_prop(sdt, node[prop].value)
                        plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()} {hex(clkprop_val)}')
                        canondef_dict.update({prop:hex(clkprop_val)})
                    elif prop == "child,required":
                        for j,child in enumerate(list(node.child_nodes.items())):
                            for k,p in enumerate(pad):
                                try:
                                    val = hex(child[1][p].value[0])
                                except KeyError:
                                    val = 0xFFFF
                                p = p.replace("-", "_")
                                p = p.replace("xlnx,", "")
                                plat.buf(f'\n#define XPAR_{label_name}_{j}_{p.upper()} {val}')
                    elif phandle_prop:
                        try:
                            prop_val = bm_config.get_phandle_regprop(sdt, prop, node[prop].value)
                            plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()} {hex(prop_val)}')
                            canondef_dict.update({prop:hex(prop_val)})
                        except KeyError:
                            pass
                    elif prop == "ranges":
                        try:
                            device_type = node['device_type'].value[0]
                            if device_type == "pci":
                                device_ispci = 1
                        except KeyError:
                            device_ispci = 0
                        if device_ispci:
                            prop_vallist = bm_config.get_pci_ranges(node, node[prop].value, pad)
                            i = 0
                            for j, prop_val in enumerate(prop_vallist):
                                if j % 2:
                                    plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()}_HIGHADDR_{i} {prop_val}')
                                    cannon_prop = prop + str("_") + str("HIGHADDR") + str("_") + str(i)
                                    canondef_dict.update({cannon_prop:prop_val})
                                    i += 1
                                else:
                                    plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()}_BASEADDR_{i} {prop_val}')
                                    cannon_prop = prop + str("_") + str("BASEADDR") + str("_") + str(i)
                                    canondef_dict.update({cannon_prop:prop_val})
                    else:
                        try:
                            prop_val = node[prop].value
                            # For boolean property if present LopperProp will return
                            # empty string convert it to baremetal config struct expected value
                            if '' in prop_val:
                                prop_val = [1]
                        except KeyError:
                            prop_val = [0]

                        if ('/bits/' in prop_val):
                            prop_val = [int(prop_val[-1][3:-1], base=16)]
                        prop = prop.replace("-", "_")
                        prop = prop.replace("xlnx,", "")
                        if len(prop_val) > 1:
                            for k,item in enumerate(prop_val):
                                cannon_prop = prop + str("_") + str(k)
                                canondef_dict.update({cannon_prop:item})
                                plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()}_{k} {item}')
                        else:
                            canondef_dict.update({prop:hex(prop_val[0])})
                            plat.buf(f'\n#define XPAR_{label_name}_{prop.upper()} {hex(prop_val[0])}')

                plat.buf(f'\n\n/* Canonical definitions for peripheral {label_name} */')
                for prop,val in sorted(canondef_dict.items(), key=lambda e: e[0][0], reverse=False):
                    plat.buf(f'\n#define XPAR_X{drv.upper()}_{index}_{prop.upper()} {val}')
                plat.buf('\n')
                                    
    # Generate Defines for Generic Nodes
    node_list = bm_config.get_mapped_nodes(sdt, node_list, options)
    prev = ""
    count = 0
    for node in node_list:
        label_name = bm_config.get_label(sdt, symbol_node, node)
        node_name = node.name
        node_name = node_name.split("@")
        node_name = node_name[0]
        if prev != node_name:
            count = 0
        else:
            count = count + 1

        prev = node_name
        label_name = label_name.upper()
        node_name = node_name.upper()
        try:
            val = bm_config.scan_reg_size(node, node['reg'].value, 0)
            plat.buf(f'\n/* Definitions for peripheral {label_name} */')
            plat.buf(f'\n#define XPAR_{label_name}_BASEADDR {hex(val[0])}\n')
            plat.buf(f'#define XPAR_{label_name}_HIGHADDR {hex(val[0] + val[1] - 1)}\n')
            temp_label = label_name.rsplit("_", 1)
            temp_label = temp_label[0]
            if temp_label != node_name:
                plat.buf(f'\n/* Canonical definitions for peripheral {label_name} */')
                node_name = node_name.replace("-", "_")
                plat.buf(f'\n#define XPAR_{node_name}_{count}_BASEADDR {hex(val[0])}\n')
                plat.buf(f'#define XPAR_{node_name}_{count}_HIGHADDR {hex(val[0] + val[1] - 1)}\n')
        except KeyError:
            pass

    # Define for Board
    board_list = ["zcu102", "kc705", "kcu105"]
    root_compat = sdt.tree['/']['compatible'].value
    board_name = [b for b in board_list for board in root_compat if b in board]
    if board_name:
        plat.buf(f'\n\n#define XPS_BOARD_{board_name[0].upper()}\n')
    
    # Memory Node related defines
    mem_ranges = get_memranges(tgt_node, sdt, options)
    for key, value in sorted(mem_ranges.items(), key=lambda e: e[1][1], reverse=True):
        start,size = value[0], value[1]
        plat.buf(f"\n#define XPAR_{key.upper()}_BASEADDRESS {hex(start)}")
        plat.buf(f"\n#define XPAR_{key.upper()}_HIGHADDRESS {hex(start + size)}")

    if cci_en:
        plat.buf("\n#define XPAR_CACHE_COHERENT \n")

    #CPU Freq related defines
    match_cpunode = bm_config.get_cpu_node(sdt, options)
    if re.search("microblaze", match_cpunode['compatible'].value[0]):
        try:
            cpu_freq = match_cpunode['xlnx,freq'].value[0]
            plat.buf(f'\n#define XPAR_CPU_CORE_CLOCK_FREQ_HZ {cpu_freq}\n')
        except KeyError:
            pass
    else:
        try:
            cpu_freq = match_cpunode['xlnx,cpu-clk-freq-hz'].value[0]
            plat.buf(f'\n\n#define XPAR_CPU_CORE_CLOCK_FREQ_HZ {cpu_freq}\n')
            pss_ref = match_cpunode['xlnx,pss-ref-clk-freq'].value[0]
            plat.buf(f'#define XPAR_PSU_PSS_REF_CLK_FREQ_HZ {pss_ref}\n')
            timestamp_clk = match_cpunode['xlnx,timestamp-clk-freq'].value[0]
            plat.buf(f'#define XPAR_CPU_TIMESTAMP_CLK_FREQ {timestamp_clk}\n')
        except KeyError:
            pass

    #Defines for STDOUT and STDIN Baseaddress
    if chosen_node:
        stdin_node = bm_config.get_stdin(sdt, chosen_node, total_nodes)
        val, size = bm_config.scan_reg_size(stdin_node, stdin_node['reg'].value, 0)
        plat.buf(f"\n#define STDOUT_BASEADDRESS {hex(val)}")
        plat.buf(f"\n#define STDIN_BASEADDRESS {hex(val)}\n")

    #Define for NUMBER_OF_SLRS
    if sdt.tree[tgt_node].propval('slrcount') != ['']:
        val = sdt.tree[tgt_node].propval('slrcount', list)[0]
        plat.buf(f"\n/* Number of SLRs */")
        plat.buf(f"\n#define NUMBER_OF_SLRS {hex(val)}\n")

    plat.buf('\n#endif  /* end of protection macro */')
    plat.out(''.join(plat.get_buf()))

    return True



from lopper.tree import *

KERNEL_FLAG = 'openamp-xlnx-kernel'
HOST_FLAG = 'openamp-host'

zynqmp_ipi_to_irq_vect_id = {
  0xff330000 : 62,
  0xFF340000 : 63,
  0xFF350000 : 64,
  0xFF310000 : 65,
  0xff320000 : 66,
  0xff380000 : 67,
}

versal_ipi_to_irq_vect_id = {
    0xff330000 : 62,
    0xff340000 : 63,
    0xFF350000 : 64,
    0xFF360000 : 65,
    0xFF370000 : 66,
    0xFF380000 : 67,
}

ipi_to_agent = {
  0xff330000 : 0x400,
  0xff340000 : 0x600,
  0xff350000 : 0x800,
  0xff360000 : 0xa00,
  0xff370000 : 0xc00,
  0xff380000 : 0xe00,
}

openamp_supported_ipis = [ 0xff330000, 0xff340000, 0xff350000, 0xff360000, 0xff370000, 0xff380000 ]

ipi_addr_to_agent = {
    0xff310000 : 'psm',
    0xff320000 : 'pmc',

    0xff330000 : 'ipi0',
    0xff340000 : 'ipi1',
    0xff350000 : 'ipi2',
    0xff360000 : 'ipi3',
    0xff370000 : 'ipi4',
    0xff380000 : 'ipi5',

}


ipi_msg_buf_dest_agent_request_offsets = {
    0xff310000 : 0x000,
    0xff320000 : 0x040,
    0xff330000 : 0x080,
    0xff340000 : 0x0c0,
    0xff350000 : 0x100,
    0xff360000 : 0x140,
    0xff370000 : 0x180,
    0xff380000 : 0x1c0,

}

source_agent_to_ipi = {
  0x000: 'psm',  0x100: 'psm',
  0x200: 'pmc',  0x300: 'pmc',
  0x400: 'ipi0', 0x500: 'ipi0',
  0x600: 'ipi1', 0x700: 'ipi1',
  0x800: 'ipi2', 0x900: 'ipi2',
  0xa00: 'ipi3', 0xb00: 'ipi3',
  0xc00: 'ipi4', 0xd00: 'ipi4',
  0xe00: 'ipi5', 0xf00: 'ipi5',

}
agent_to_ipi_bitmask = {
  0x000: 0x1 ,
  0x200: 0x2 ,  
  0x400: 0x4,
  0x600: 0x8,
  0x800: 0x10,
  0xa00: 0x20,
  0xc00: 0x40,
  0xe00: 0x80,

  0x100: 0x1 ,
  0x300: 0x2 ,  
  0x500: 0x4,
  0x700: 0x8,
  0x900: 0x10,
  0xb00: 0x20,
  0xd00: 0x40,
  0xf00: 0x80,

}


class SOC_TYPE:
    UNINITIALIZED = -1
    VERSAL = 0
    ZYNQMP = 1
    ZYNQ = 2

general_template="""
#ifndef OPENAMP_LOPPER_INFO_H_
#define OPENAMP_LOPPER_INFO_H_


#define CHANNEL_0_MEM_VDEV0VRING0_ADDR	{CHANNEL_0_VDEV0VRING0_ADDR}
#define CHANNEL_0_MEM_VDEV0VRING0_RANGE	{CHANNEL_0_VDEV0VRING0_RANGE}
#define CHANNEL_0_MEM_VDEV0VRING1_ADDR	{CHANNEL_0_VDEV0VRING1_ADDR}
#define CHANNEL_0_MEM_VDEV0VRING1_RANGE	{CHANNEL_0_VDEV0VRING1_RANGE}
#define CHANNEL_0_MEM_VDEV0BUFFER_ADDR	{CHANNEL_0_VDEV0BUFFER_ADDR}
#define CHANNEL_0_MEM_VDEV0BUFFER_RANGE	{CHANNEL_0_VDEV0BUFFER_RANGE}
#define CHANNEL_0_MEM_ELFLOAD_ADDR	{CHANNEL_0_ELFLOAD_ADDR}
#define CHANNEL_0_MEM_ELFLOAD_RANGE	{CHANNEL_0_ELFLOAD_RANGE}
#define CHANNEL_0_MEM_RANGE	        {CHANNEL_0_MEM_RANGE}

#define CHANNEL_0_MEM_SHARED_MEM_PA	{CHANNEL_0_SHARED_MEM_PA}
#define CHANNEL_0_MEM_RING_TX           {CHANNEL_0_TX}
#define CHANNEL_0_MEM_RING_RX	        {CHANNEL_0_RX}
#define CHANNEL_0_MEM_SHARED_MEM_SIZE	{CHANNEL_0_SHARED_MEM_SIZE}
#define CHANNEL_0_MEM_SHARED_BUF_OFFSET	{CHANNEL_0_SHARED_BUF_OFFSET}
#define CHANNEL_0_MEM_VRING_MEM_SIZE	{CHANNEL_0_VRING_MEM_SIZE}
#define CHANNEL_0_MEM_RSC_MEM_SIZE	{CHANNEL_0_RSC_MEM_SIZE}
#define CHANNEL_0_MEM_NUM_VRINGS	2
#define CHANNEL_0_MEM_VRING_ALIGN	0x1000
#define CHANNEL_0_MEM_VRING_SIZE	256
#define CHANNEL_0_MEM_NUM_TABLE_ENTRIES	1
#define REMOTE_BUS_NAME                 "generic"
#define MASTER_BUS_NAME                 "platform"

{soc_template}

#endif /* OPENAMP_LOPPER_INFO_H_ */
"""

zynq_template="""
#define DEVICE_MEMORY 0xC06     /* Device memory */
#define STRONG_ORDERED 0xC02    /* Strongly ordered */
#define RESERVED 0x0            /* reserved memory */
#define NORM_NONCACHE 0x11DE2   /* Normal Non-cacheable */
#define REMOTE_SCUGIC_DEV_NAME	"scugic_dev"
#define SCUGIC_PERIPH_BASE	0xF8F00000
#define SCUGIC_DIST_BASE	(SCUGIC_PERIPH_BASE + 0x00001000)
#define ZYNQ_CPU_ID_MASK	0x1UL
/* SGIs */
#define SGI_TO_NOTIFY           15 /* SGI to notify the remote */
#define SGI_NOTIFICATION        14 /* SGI from the remote */
"""

r5_template="""
#define CHANNEL_0_MASTER_IPI_BASE_ADDR	{MASTER_IPI_BASE_ADDR}
#define CHANNEL_0_MASTER_IPI_NAME	{MASTER_IPI_NAME}
#define CHANNEL_0_MASTER_IRQ_VECT_ID    {MASTER_IRQ_VECT_ID}
#define CHANNEL_0_MASTER_CHN_BITMASK    {MASTER_CHN_BITMASK}
#define CHANNEL_0_REMOTE_IPI_BASE_ADDR  {REMOTE_IPI_BASE_ADDR}
#define CHANNEL_0_REMOTE_IPI_NAME       {REMOTE_IPI_NAME}
#define CHANNEL_0_REMOTE_IRQ_VECT_ID    {REMOTE_IRQ_VECT_ID}
#define CHANNEL_0_REMOTE_CHN_BITMASK    {REMOTE_CHN_BITMASK}
"""


# platform determine SoC
# is_kernel_case for SoC's that support user/kernelspace
# inputs is dictionary with inputs for template
def write_openamp_header(platform, is_kernel_case, inputs, options):
    if (len(options["args"])) > 0:
        f_name = options["args"][0]
    else:
        f_name = "openamp_lopper_info.h"
    try:
        verbose = options['verbose']
    except:
        verbose = 0

    if not inputs:
        print( "[WARNING]: unable to generate openamp_lopper_info.h, no valid inputs" )
        return

    f =  open(f_name,"w")
    if platform == SOC_TYPE.ZYNQ:
        inputs["soc_template"] = zynq_template
    else:
        inputs["soc_template"] = r5_template.format(**inputs)
        
    f.write(general_template.format(**inputs))
    f.close()

def generate_openamp_file(carveout_list, options, platform, is_kernel_case, inputs):
    symbol_name = "CHANNEL_0_MEM_"
    current_channel_count = 0 # if == 4 then got complete channel range
    vring_mems = []
    channel_range = 0

    if platform == SOC_TYPE.ZYNQ:
        addr_column = 0
        range_column = 1
    else:
        addr_column = 1
        range_column = 3

    for i in carveout_list:

        if "vdev0buffer" in i[0]:
            current_channel_count += 1
            inputs["CHANNEL_0_VDEV0BUFFER_ADDR"] = i[1][addr_column]
            inputs["CHANNEL_0_VDEV0BUFFER_RANGE"] = i[1][range_column]
            inputs["CHANNEL_0_SHARED_MEM_SIZE"] = i[1][range_column]
            vdev0buffer_range = i[1][range_column]
            channel_range += int(i[1][range_column],16)
        elif "vdev0vring0" in i[0]:
            current_channel_count += 1
            inputs["CHANNEL_0_VDEV0VRING0_ADDR"] = i[1][addr_column]
            inputs["CHANNEL_0_VDEV0VRING0_RANGE"] = i[1][range_column]
            inputs["CHANNEL_0_SHARED_MEM_PA"] =  i[1][range_column]
            if is_kernel_case == False:
                inputs["CHANNEL_0_TX"] = i[1][addr_column]

            channel_range += int(i[1][range_column],16)
            vring_mems.append(i[1][range_column])
        elif "vdev0vring1" in i[0]:
            vring_mems.append(i[1][range_column])
            inputs["CHANNEL_0_VDEV0VRING1_ADDR"] = i[1][addr_column]
            inputs["CHANNEL_0_VDEV0VRING1_RANGE"] = i[1][range_column]
            if is_kernel_case == False:
                inputs["CHANNEL_0_RX"] = i[1][addr_column]

            current_channel_count += 1
            channel_range += int(i[1][range_column],16)
        elif "elfload" in i[0]:
            current_channel_count += 1
            channel_range += int(i[1][range_column],16)
            inputs["CHANNEL_0_ELFLOAD_ADDR"] = i[1][addr_column]
            inputs["CHANNEL_0_ELFLOAD_RANGE"] = i[1][range_column]

        if current_channel_count == 4:
            inputs["CHANNEL_0_MEM_RANGE"] = hex(channel_range)
            current_channel_count += 1 # TODO account for >1 channels
            vring_mems_size_total = 0
            for i in vring_mems:
                vring_mems_size_total += int(i,16)
            inputs["CHANNEL_0_SHARED_BUF_OFFSET"] = hex(vring_mems_size_total)
            inputs["CHANNEL_0_VRING_MEM_SIZE"] = hex(vring_mems_size_total)

            vring_mem_size = 0

    write_openamp_header(platform, is_kernel_case, inputs, options)

def parse_memory_carevouts(sdt, options, remoteproc_node):
    try:
        verbose = options['verbose']
    except:
        verbose = 0

    carveout_list = [] # string representation of mem carveout nodes
    phandle_list = []

    reserved_mem_node = sdt.tree["/reserved-memory"]

    for node in reserved_mem_node.subnodes():
        if node.props("compatible") != [] and "openamp,xlnx,mem-carveout" in node["compatible"].value:
            phandle_list.append(node.phandle)
            carveout_list.append( ( (str(node), str(node['reg']).replace("reg = <","").replace(">;","").split(" ")) ))

    # output to DT
    if remoteproc_node != None:
        try:
            remoteproc_node["memory-region"].value = phandle_list
            remoteproc_node.sync ( sdt.FDT )
        except:
            if verbose > 0:
                print( "[ERROR]: cannot find the target remoteproc node ")

    return carveout_list

def resolve_remoteproc_carveouts( tree, subnode, verbose = 0 ):
    prop = None
    domain_node = None
    new_prop_val = []

    if subnode.props("elfload") == []:
        print("WARNING:", "remoteproc relation does not have elfload carveouts", subnode.abs_path)
        return False

    prop = subnode.props("elfload")[0]

    # for each carveout
    # look for it in domain's reserved memory
    # get or create phandle of the carveout
    for carveout_str in prop.value:
        current_node = None

        for n in subnode.tree["/"].subnodes():
            if carveout_str == n.name or carveout_str == n.label:
                current_node = n
                break

        if current_node == None:
            print("WARNING:", "rpmsg relation can't find carveout name", carveout_str)
            return False

        if current_node.phandle == 0:
            current_node.phandle_or_create()

        if current_node.props("phandle") == []:
            current_node + LopperProp(name="phandle", value=current_node.phandle)

        new_prop_val.append(current_node.phandle)

    # update value of property to have phandles
    prop.value = new_prop_val

    return True

def resolve_rpmsg_carveouts( tree, subnode, verbose = 0 ):
    prop = None
    res_mem_node = None
    new_prop_val = []

    if subnode.props("carveouts") == []:
        print("WARNING:", "rpmsg relation does not have carveouts")
        return False

    prop = subnode.props("carveouts")[0]

    # for each carveout
    # look for it in domain's reserved memory
    # get or create phandle of the carveout

    try:
        res_mem_node = subnode.tree[subnode.parent.parent.abs_path + "/reserved-memory"]
    except KeyError:
        print("WARNING:", "rpmsg relation's backing domain does not have reserved memory with carveouts.", subnode.abs_path)
        return False

    for carveout_str in prop.value:
        current_node = None

        for n in res_mem_node.subnodes():
            if carveout_str == n.name or carveout_str == n.label:
                current_node = n
                break

        if current_node == None:
            print("WARNING:", "rpmsg relation can't find carveout name", carveout_str)
            return False

        if current_node.phandle == 0:
            current_node.phandle_or_create()

        if current_node.props("phandle") == []:
            current_node + LopperProp(name="phandle", value=current_node.phandle)

        new_prop_val.append(current_node.phandle)

    # update value of property to have phandles
    prop.value = new_prop_val

    return True

def resolve_rpmsg_mbox( tree, subnode, verbose = 0 ):
    mbox_node = None
    prop = None

    if subnode.props("mbox") == []:
        print("WARNING:", "rpmsg relation does not have mbox")
        return False

    prop = subnode.props("mbox")[0]
    search_str = prop.value.strip()

    for n in subnode.tree["/"].subnodes():
        if search_str == n.name or search_str == n.label:
            mbox_node = n
            break

    if mbox_node == None:
        print("resolve_rpmsg_mbox: ", tree.lnodes(n.name, exact = False) )

    if mbox_node == None or mbox_node == []:
        print("WARNING:", "rpmsg relation can't find mbox name: ", prop.value)
        return False

    if mbox_node.phandle == 0:
        mbox_node.phandle_or_create()

    prop.value = mbox_node.phandle
    if mbox_node.props("phandle") == []:
        mbox_node + LopperProp(name="phandle", value=mbox_node.phandle)
    return True

def resolve_host_remote( tree, subnode, verbose = 0 ):
    prop = None
    domain_node = None

    if subnode.props("host") != [] and subnode.props("remote") != []:
        print("WARNING:", "relation has both host and remote")
        return False
    elif subnode.props("host") != []:
        prop = subnode.props("host")[0]
    else:
        prop = subnode.props("remote")[0]


    for n in tree["/domains"].subnodes():
        if prop.value in n.name:
            domain_node = n
            break

    if domain_node.phandle == 0:
        domain_node.phandle_or_create()

    prop.value = domain_node.phandle
    if domain_node.props("phandle") == []:
        domain_node + LopperProp(name="phandle", value=domain_node.phandle)
    return True



#/*
# * Copyright (c) 2019,2020 Xilinx Inc. All rights reserved.
# *
# * Author:
# *       Bruce Ashfield <bruce.ashfield@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import copy
import struct
import sys
import types
import unittest
import os
import getopt
import re
import subprocess
import shutil
from pathlib import Path
from pathlib import PurePath
from io import StringIO
import contextlib
import importlib
from lopper import Lopper
from lopper import LopperFmt
import lopper
from lopper.tree import *
from re import *
from openamp_xlnx_common import *


def is_compat( node, compat_string_to_test ):
    if re.search( "openamp,xlnx-zynq-a9", compat_string_to_test):
        return xlnx_openamp_zynq
    return ""

# tgt_node: is the openamp domain node number
# sdt: is the system device tree
# TODO: this routine needs to be factored and made smaller
def xlnx_openamp_zynq( tgt_node, sdt, options ):
    try:
        verbose = options['verbose']
    except:
        verbose = 0

    if verbose:
        print( "[INFO]: cb: xlnx_openamp_zynq( %s, %s, %s )" % (tgt_node, sdt, verbose))

    domain_node = sdt.tree[tgt_node]

    root_node = sdt.tree["/"]
    try:
        memory_node = sdt.tree[ "/reserved-memory" ]
        is_kernel_case = True
    except:
        return False

    remoteproc_node = sdt.tree["/remoteproc0"]
    mem_carveouts = parse_memory_carevouts(sdt, options, remoteproc_node)
    # last arg (True) denotes for kernelspace case
    inputs = {
        "CHANNEL_0_MEM_SIZE" : "0x80000UL",
        "CHANNEL_0_SHARED_BUF_OFFSET" : "0x80000UL",
        "CHANNEL_0_RSC_MEM_SIZE" : "0x2000UL",
        "CHANNEL_0_TX" : "FW_RSC_U32_ADDR_ANY",
        "CHANNEL_0_RX" : "FW_RSC_U32_ADDR_ANY",
        "CHANNEL_0_VRING_MEM_SIZE" : "0x8000",
    }

    generate_openamp_file( mem_carveouts, options, SOC_TYPE.ZYNQ, is_kernel_case, inputs )
    return True



#/*
# * Copyright (c) 2021 Xilinx Inc. All rights reserved.
# *
# * Author:
# *       Appana Durga Kedareswara rao <appana.durga.rao@xilinx.com>
# *       Naga Sureshkumar Relli <naga.sureshkumar.relli@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import struct
import sys
import types
import os
import getopt
import re
import subprocess
from pathlib import Path
from pathlib import PurePath
from lopper import Lopper
from lopper import LopperFmt
import lopper
from lopper_tree import *
from lopper_tree import LopperNode
from re import *
import yaml
import glob
from collections import OrderedDict

sys.path.append(os.path.dirname(__file__))
from baremetalconfig_xlnx import *

def is_compat( node, compat_string_to_test ):
    if re.search( "module,xlnx_overlay_dt", compat_string_to_test):
        return xlnx_generate_overlay_dt
    return ""

def get_label(sdt, symbol_node, node):
    prop_dict = Lopper.node_properties_as_dict(sdt.FDT, symbol_node.abs_path, False)
    match = [label for label,node_abs in prop_dict.items() if re.match(node_abs[0], node.abs_path) and len(node_abs[0]) == len(node.abs_path)]
    if match:
        return match[0]
    else:
        return None

def usage():
    prog = os.path.basename(sys.argv[0])
    print('Usage: %s <system device tree> -- <xlnx_overlay_dt.py> <machine name> <configuration>' % prog)
    print('  machine name:         cortexa53-zynqmp or cortexa72-versal')
    print('  configuration:        should be \'full\'' )

"""
This API generates the overlay dts file by taking pl.dtsi
generated from DTG++.
Args:
    tgt_node: is the baremetal config top level domain node number
    sdt:      is the system device-tree
    options:  There are two valid options
              Machine name as cortexa53-zynqmp or cortexa72-versal
              An optional argument full/partial
                  The default will be full
"""
def xlnx_generate_overlay_dt(tgt_node, sdt, options):
    root_node = sdt.tree[tgt_node]
    root_sub_nodes = root_node.subnodes()

    symbol_node = ""
    plat = DtbtoCStruct('pl.dtsi')
    for node in root_sub_nodes:
        try:
            if node.name == "__symbols__":
                symbol_node = node
        except:
           pass

    # Get the list of nodes that are added to fragment@1
    # i,e fragment@1/overlay@1 node. we can ignore these nodes
    # while reading the tree to add nodes under fragment@2/overlay@2
    ignore_list = []
    for node in root_sub_nodes:

        try:
            if re.search("afi0" , node.name) or re.search("clocking" , node.name):
               ignore_list.append(node)
               
        except:
           pass

    # Initialize the variables to its defaults
    root = 1
    tab_len = 0
    platform = "None"
    config = "None"
    external_flag = 1
    parent_node = ""
    parent_tab = 0
    try:
        platform = options['args'][0]
        if options['args'][1] == "external_fpga":
            external_flag = options['args'][1]
        else:
            config = options['args'][1]
            external_flag = options['args'][2]
    except:
        pass

    # If no config option is provided then the default is
    # full bit stream support
    if config == "None":
        config = "full"

    if config != "full":
        print('%s is not a valid argument' % str(config))
        usage()
        sys.exit(1)        

    for node in root_sub_nodes:
        set_ignore = 0

        try:
            path = node.abs_path
            ret = path.split('/')
            pl = 0
            rt = len(ret)
            child_len = len(node.child_nodes)

            # Only check for the nodes under amba_pl
            for x in ret:
                if x == "amba_pl":
                   pl = 1

            if pl == 1:
                if child_len != 0 and node.name != "amba_pl":
                    if child_len > 1:
                        tab_len = 0
                    else:
                        tab_len = tab_len + int(child_len)

                if root == 1:
                    # Create overlay0: __overlay__ node as first node under fragment@0
                    plat.buf('/dts-v1/;')
                    plat.buf('\n/plugin/;')
                    plat.buf('\n/ {')
                    plat.buf('\n\tfragment@0 {')
                    if platform == "cortexa53-zynqmp":
                        plat.buf('\n\t\ttarget = <&fpga_full>;')
                    else:
                        plat.buf('\n\t\ttarget = <&fpga>;')
                    plat.buf('\n\t\toverlay0: __overlay__ {')
                    plat.buf('\n\t\t\t#address-cells = <2>;')
                    plat.buf('\n\t\t\t#size-cells = <2>;')
                    try:
                       if config == "full":
                           if platform == "cortexa53-zynqmp":
                               plat.buf('\n\t\t\t%s' % node['firmware-name'])
                           elif platform == "cortexa72-versal":
                               if external_flag == "external_fpga":
                                   plat.buf('\n\t\t\texternal-fpga-config;')
                               else:
                                   plat.buf('\n\t\t\t%s' % node['firmware-name'])
                           else:
                               print('%s is not a valid Machine' % str(platform))
                               sys.exit(1)        
                               
                    except:
                        pass

                    plat.buf('\n\t\t};')
                    plat.buf('\n\t};')

                    # Create overlay1: __overlay__ node under fragment@1
                    plat.buf('\n\tfragment@1 {')
                    plat.buf('\n\t\ttarget = <&amba>;')
                    plat.buf('\n\t\toverlay1: __overlay__ {')
                    plat.buf('\n\t\t\t#address-cells = <2>;')
                    plat.buf('\n\t\t\t#size-cells = <2>;')

                    # Add afi and clocking nodes to fragment@1     
                    if platform == "cortexa53-zynqmp":
                        for inode in ignore_list:
                            label_name = get_label(sdt, symbol_node, inode)
                            plat.buf('\n\t\t\t%s: %s {' % (label_name, inode.name))
                            for p in inode.__props__.values():
                                if re.search("phandle =", str(p)) or str(p) == '':
                                    continue
                                plat.buf('\n\t\t\t\t%s' % p)
                            plat.buf('\n\t\t\t};')
                        plat.buf('\n\t\t};')
                        plat.buf('\n\t};')
                    
                        # Create overlaye2: __overlay__ node under fragment@2
                        plat.buf('\n\tfragment@2 {')
                        plat.buf('\n\t\ttarget = <&amba>;')
                        plat.buf('\n\t\toverlay2: __overlay__ {')
                        plat.buf('\n\t\t\t#address-cells = <2>;')
                        plat.buf('\n\t\t\t#size-cells = <2>;')

                    root = 0

                # Now add all the nodes except the nodes that are added
                # that are added under fragment@1
                for ignoreip in ignore_list:
                    if re.match(ignoreip.name , node.name):
                       set_ignore = 1

                if set_ignore == 1:
                    continue

                # Add all the nodes exists under amba_pl to fragment@2
                if node.name != "amba_pl":
                    if parent_tab == 1 and parent_node == node.parent:
                        plat.buf('\n')
                        plat.buf('\t' * int(rt))
                        plat.buf('};')
                        parent_tab = 0

                    label_name = get_label(sdt, symbol_node, node)
                    plat.buf('\n')
                    rt = int(rt+1)
                    plat.buf('\t' * int(rt-1))
                    plat.buf('%s: %s {' % (label_name, node.name))
                    for p in node.__props__.values():
                        if re.search("phandle =", str(p)) or str(p) == '':
                            continue
                        plat.buf('\n')
                        plat.buf('\t' * int(rt))
                        if re.search("clocks =", str(p)):
                            plat.buf('%s' % node['clocks'])
                        else:
                            plat.buf('%s' % p)

                    plat.buf('\n')
                    plat.buf('\t' * int(rt-1))
                    if child_len < 1:
                        plat.buf('};')
                        for count in range(tab_len):
                            plat.buf('\n')
                            plat.buf('\t' * int(int(rt-2)-count))
                            plat.buf('};')

                        tab_len = 0
                    else:
                        if child_len > tab_len:
                            parent_node = node.parent
                            parent_tab = 1
        except:
           pass

    plat.buf('\n\t\t};')
    plat.buf('\n\t};')
    plat.buf('\n};')
    plat.out(''.join(plat.get_buf()))

    return True



#/*
# * Copyright (c) 2021 Xilinx Inc. All rights reserved.
# *
# * Author:
# *       Appana Durga Kedareswara rao <appana.durga.rao@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import sys
import os
import re
import shutil
import glob

sys.path.append(os.path.dirname(__file__))
from baremetalconfig_xlnx import get_mapped_nodes, get_label, compat_list
import common_utils as utils

def is_compat( node, compat_string_to_test ):
    if re.search( "module,baremetal_gentestapp_xlnx", compat_string_to_test):
        return xlnx_generate_testapp
    return ""

# tgt_node: is the baremetal config top level domain node number
# sdt: is the system device-tree
# options: baremetal application source path
def xlnx_generate_testapp(tgt_node, sdt, options):
    root_node = sdt.tree[tgt_node]
    root_sub_nodes = root_node.subnodes()
    node_list = []
    # Traverse the tree and find the nodes having status=ok property
    # and create a compatible_list from these nodes.
    symbol_node = ""
    for node in root_sub_nodes:
        try:
            if node.name == "__symbols__":
                symbol_node = node
            status = node["status"].value
            if "okay" in status:
                node_list.append(node)
        except:
           pass

    node_list = get_mapped_nodes(sdt, node_list, options)

    test_file_string = f'''
#include <stdio.h>
#include "xparameters.h"
#include "xil_printf.h"
'''
    periph_file_list = []

    repo_path_data = options['args'][1]
    yaml_file_list = []

    if utils.is_file(repo_path_data):
        repo_schema = utils.load_yaml(repo_path_data)
        drv_data = repo_schema['driver']
        for entries in drv_data.keys():
            drv_path_in_yaml = drv_data[entries]['vless']
            drv_name_in_yaml = os.path.basename(drv_path_in_yaml)
            yaml_file_list += [os.path.join(drv_path_in_yaml, 'data', f"{drv_name_in_yaml}.yaml")]
    else:
        drv_dir = os.path.join(repo_path_data, "XilinxProcessorIPLib", "drivers")
        if utils.is_dir(drv_dir):
            yaml_file_list = glob.glob(drv_dir + '/**/data/*.yaml', recursive=True)

    testapp_data = {}
    testapp_name = {}
    for yaml_file in yaml_file_list:
        schema = utils.load_yaml(yaml_file)
        driver_compatlist = compat_list(schema)
        driver_nodes = []
        drv_name = utils.get_base_name(yaml_file).replace('.yaml','')
        drv_data_path = utils.get_dir_path(yaml_file)
        drv_path = utils.get_dir_path(drv_data_path)
        try:
            drv_config_name = schema['config']
            drv_config_name = drv_config_name[0].rsplit("_", 1)[-2]
        except KeyError:
            drv_config_name = drv_name
        try:
            testapp_schema = schema['tapp']
            tapp_drv_header_file_name = f"{drv_name}_header.h"
            tapp_drv_header_src_file = os.path.join(drv_data_path, tapp_drv_header_file_name)
            tapp_drv_header_dst_file = os.path.join(sdt.outdir, tapp_drv_header_file_name)
            periph_file_list += [tapp_drv_header_dst_file]
            utils.copy_file(tapp_drv_header_src_file, tapp_drv_header_dst_file)

            test_file_string += f'''
#include "x{drv_name}.h"
#include "{tapp_drv_header_file_name}"
'''
            with open(tapp_drv_header_dst_file, 'r+') as fd:
                content = fd.readlines()
                content.insert(0, "#define TESTAPP_GEN\n")
                fd.seek(0, 0)
                fd.writelines(content)

            for compat in driver_compatlist:
                for node in node_list:
                    compat_string = node['compatible'].value[0]
                    label_name = get_label(sdt, symbol_node, node)
                    if compat in compat_string:
                        driver_nodes.append(node)
                        dec = []
                        for app,prop in testapp_schema.items():
                            example_dir = os.path.join(drv_path, "examples")
                            example_file_src_path = os.path.join(example_dir, app)
                            example_file_dst_path = os.path.join(sdt.outdir, app)
                            try:
                                has_hwdep = testapp_schema[app]['hwproperties'][0]
                                try:
                                    val = node[has_hwdep].value
                                    has_hwdep = 0
                                except KeyError:
                                    has_hwdep = 1
                            except KeyError:
                                has_hwdep = 0

                            if not has_hwdep and utils.is_file(example_file_src_path):
                                periph_file_list += [example_file_dst_path]
                                utils.copy_file(example_file_src_path, example_file_dst_path)
                                with open(example_file_dst_path, 'r+') as fd:
                                    content = fd.readlines()
                                    content.insert(0, "#define TESTAPP_GEN\n")
                                    fd.seek(0, 0)
                                    fd.writelines(content)
                                dec.append(testapp_schema[app]['declaration'])
                        testapp_data.update({label_name:dec})
                        testapp_name.update({label_name:drv_config_name})
        except KeyError:
            testapp_schema = {}


    test_file_string += f'''int main ()
{{
'''
    for node,drv_config_name in testapp_name.items():
        test_file_string += f'''
    static {drv_config_name} {node};'''

    test_file_string += f'''

    print("---Entering main---\\n\\r");
'''

    for node, testapp in testapp_data.items():
        xpar_def = f"XPAR_{node.upper()}_BASEADDR"

        for app in testapp:
            if 'SelfTest' in app:
                status_assignment = f"status = {app}({xpar_def});"
            else:
                status_assignment = f"status = {app}(&{node}, {xpar_def});"

            test_file_string += f'''
    {{
        int status;
        print("\\r\\nRunning {app} for {node} ... \\r\\n");
        {status_assignment}
        if (status == 0) {{
            print("{app} PASSED \\r\\n");
        }} else {{
            print("{app} FAILED \\r\\n");
        }}
    }}
'''

    test_file_string += f'''
    print("---Exiting main---");
    return 0;
}}
'''

    testperiph_file = os.path.join(sdt.outdir, 'testperiph.c')
    with open(testperiph_file, 'w') as file_handle:
        file_handle.writelines(test_file_string)
        periph_file_list += [testperiph_file]

    with open(os.path.join(sdt.outdir, 'file_list.txt'), 'w') as file_handle:
        periph_file_list = set(periph_file_list)
        for entry in periph_file_list:
            file_handle.writelines(f'{entry}\n')

    return True


#/*
# * Copyright (c) 2019,2020 Xilinx Inc. All rights reserved.
# *
# * Author:
# *       Bruce Ashfield <bruce.ashfield@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import struct
import sys
import types
import unittest
import os
import getopt
import re
import subprocess
import shutil
from pathlib import Path
from pathlib import PurePath
from io import StringIO
import contextlib
import importlib
from lopper import Lopper
from lopper import LopperFmt
from lopper.tree import LopperAction
import lopper
import lopper_lib
from itertools import chain

def is_compat( node, compat_string_to_test ):
    if re.search( "access-domain,domain-v1", compat_string_to_test):
        return core_domain_access
    return ""

# tests for a bit that is set, going fro 31 -> 0 from MSB to LSB
def check_bit_set(n, k):
    if n & (1 << (k)):
        return True

    return False

# tree: is the lopper system device-tree
def domain_get_subnodes(tree):
    try:
        domain_node = tree['/domains']
    except:
        domain_node = None

    direct_node_refs = []

    if domain_node:
        for node in domain_node.subnodes():
            # 1) memory access/node = <> nodes
            try:
                mem_node = node["memory"].value
                direct_node_refs.append( node )
            except:
                pass
            # 2) direct access = <> nodes
            a_nodes = lopper_lib.node_accesses( tree, node )
            if a_nodes:
                direct_node_refs.append( node )
            # 3) include = <> nodes
            try:
                i_nodes = lopper_lib.includes( tree, node['include'])
                if i_nodes:
                    direct_node_refs.append( node )
            except:
                pass

    # Remove duplicate entries
    direct_node_refs = list(dict.fromkeys(direct_node_refs))
    return direct_node_refs

# node: is the domain node number
# mem_val: Memory node address and size value to be updated
# This api takes the memory value(address and size) and creates
# a new memory node value for memory node reg property based on the
# address-cells and size-cells property.
def update_mem_node(node, mem_val):
    ac = node.parent['#address-cells'][0]
    sc = node.parent['#size-cells'][0]

    new_mem_val = []
    mem_reg_pairs = len(mem_val)/2
    addr_list = mem_val[::2]
    size_list = mem_val[1::2]
    for i in range(0, int(mem_reg_pairs)):
        for j in range(0, ac):
            high_addr = 0
            val = str(hex(addr_list[i]))[2:]
            if len(val) > 8:
                high_addr = 1
            if j == ac-1:
                if len(val) > 8:
                    pad = len(val) - 8
                    upper_val = val[:pad]
                    lower_val = val[pad:]
                    new_mem_val.append(int(upper_val, base=16))
                    new_mem_val.append(int(lower_val, base=16))
                else:
                     new_mem_val.append(addr_list[i])
            elif high_addr != 1:
                new_mem_val.append(0)
        for j in range(0, sc):
            if j == sc-1:
                new_mem_val.append(size_list[i])
            else:
                new_mem_val.append(0)
    return new_mem_val

# tgt_node: is the domain node number
# sdt: is the system device tree
def core_domain_access( tgt_node, sdt, options ):
    try:
        verbose = options['verbose']
    except:
        verbose = 0

    # reset the treewide ref counting
    sdt.tree.ref = 0
    domain_node = sdt.tree[tgt_node]

    if verbose:
        print( "[INFO]: cb: core_domain_access( %s, %s, %s )" % (domain_node, sdt, verbose))

    direct_node_refs = []

    # 1) direct access = <> nodes
    a_nodes = lopper_lib.node_accesses( sdt.tree, domain_node )
    for anode in a_nodes:
        # add a refcount to the node and it's parents
        sdt.tree.ref_all( anode, True )
        direct_node_refs.append( anode )

    # 2) are there resource group includes ?, they can have access = <> as well
    try:
        includes = domain_node["include"]
    except:
        includes = None

    if includes:
        include_nodes = lopper_lib.includes( sdt.tree, domain_node["include"] )

        for i in include_nodes:
            a_nodes = lopper_lib.node_accesses( sdt.tree, i )
            for anode in a_nodes:
                sdt.tree.ref_all( anode, True )
                direct_node_refs.append( anode )

    # 3) cpus access
    try:
        cpu_prop = domain_node['cpus']
    except:
        if verbose:
            print( "[WARNING]: core_domain: domain node does not have a cpu link" )
        cpu_prop = None

    if cpu_prop:
        refd_cpus, unrefd_cpus = lopper_lib.cpu_refs( sdt.tree, cpu_prop, verbose )
        if refd_cpus:
            ref_nodes = sdt.tree.refd( "/cpus.*/cpu.*" )

            # now we do two types of refcount delete
            #   - betweenthe cpu clusters
            #   - on the cpus within a cluster

            # The following filter code will check for nodes that are compatible to
            # cpus,cluster and if they haven't been referenced, delete them.
            xform_path = "/"
            prop = "cpus,cluster"
            code = """
                     p = node.propval('compatible')
                     if p and "{0}" in p:
                         if node.ref <= 0:
                             return True

                     return False
                   """.format( prop )

            if verbose:
                print( "[INFO]: core_domain_access: filtering on:\n------%s\n-------\n" % code )

            # the action will be taken if the code block returns 'true'
            # Lopper.node_filter( sdt, xform_path, LopperAction.DELETE, code, verbose )
            sdt.tree.filter( xform_path, LopperAction.DELETE, code, None, verbose )

            for s in unrefd_cpus:
                try:
                    if verbose:
                        print( "[INFO]: core_domain_access: deleting unrefernced subcpu: %s" % s.abs_path )
                    sdt.tree.delete( s )
                except Exception as e:
                    print( "[WARNING]: %s" % e )

    # 4) directly accessed nodes. Check their type. If they are busses,
    #    we have some sedoncary processing to do.
    nodes_to_filter = []
    for anode in direct_node_refs:
        node_types = lopper_lib.node_ancestor_types( anode )
        simple_bus = lopper_lib.node_ancestors_of_type( anode, "simple-bus" )
        if simple_bus:
            for i,s in enumerate(simple_bus):
                if not s in nodes_to_filter:
                    if verbose > 1:
                        print( "[INFO]: core_domain_access: rsimple bus processing for: %s" % anode.name )

                    nodes_to_filter.append( s )

        reserved_memory = "reserved-memory" in chain(*node_types)
        if reserved_memory:
            if verbose > 1:
                print( "[INFO]: core_domain_access: reserved memory processing for: %s" % anode.name )
            nodes_to_filter.append( anode.parent )

    # 5) filter nodes that don't have refcounts
    #
    # filter #1:
    #    - starting at /
    #    - drop any unreferenced nodes that are of type simple-bus
    prop = "simple-bus"
    code = """
           p = node.propval( 'compatible' )
           if p and "{0}" in p:
               r = node.ref
               if r <= 0:
                   return True
               else:
                   return False
           else:
               return False
           """.format( prop )

    if verbose:
        print( "[INFO]: core_domain_access: filtering on:\n------%s\n-------\n" % code )

    sdt.tree.filter( "/", LopperAction.DELETE, code, None, verbose )

    # filter #2:
    #    - starting at simple-bus nodes
    #    - drop any unreferenced elements

    # filter #3:
    #    - starting at reserved memory parent
    #    - drop any unreferenced elements

    for n in nodes_to_filter:
        code = """
               p = node.ref
               if p <= 0:
                   return True
               else:
                   return False
               """
        if verbose:
            print( "[INFO]: core_domain_access: filtering on:\n------%s\n-------\n" % code )

        sdt.tree.filter( n + "/", LopperAction.DELETE, code, None, verbose )


    # 6) memory node processing
    try:
        memory_int = domain_node['memory'].int()
        memory_hex = domain_node['memory'].hex()
    except Exception as e:
        memory_hex = 0x0
        memory_int = 0

    # This may be moved to the top of the domain process and then when we are
    # processing cpus and bus nodes, we can apply the memory to ranges <>, etc,
    # and modify them accordingly.
    if verbose > 1:
        print( "[INFO]: core_domain_access: memory property: %s" % memory_hex )

    # 1) find if there's a top level memory node
    try:
        memory_node = sdt.tree["/memory@.*"]
    except:
        memory_node = None

    if memory_node:
        if verbose:
            print( "[INFO]: core_domain_access: memory node found (%s), modifying to match domain memory" % memory_node )

        # 2) modify that memory property to match the node we have here
        # memprop_old = sdt.FDT.getprop(memory_node, 'reg' )
        # num_bits = len(memprop_old)
        # a = 0
        # b = 1
        # c = 0
        # d = 1
        # val = a.to_bytes(4,byteorder='big') + b.to_bytes(4,byteorder='big') + c.to_bytes(4,byteorder='big') + d.to_bytes(4,byteorder='big')

        # TODO: this seems to be shortening up the system memory node. Check to see
        #       if the openamp node is being propery interpreted

        try:
            memory_node['reg'].value = memory_int
        except:
            pass

    # final) deal with unreferenced nodes
    refd_nodes = sdt.tree.refd()
    if verbose:
        for p in refd_nodes:
            code = """
                p = node.ref
                if p <= 0:
                    return True
                else:
                    return False
                """
            # delete any unreferenced nodes
            # not currently enabled, as it deletes our domain and other nodes
            # of values. We could refernece those nodes explicitly if we want
            # to use this as a final house cleaning step in the future.
            # sdt.tree.filter( "/", LopperAction.DELETE, code, None, verbose )


    #sys.exit(1)

    return True



#/*
# * Copyright (c) 2020 Xilinx Inc. All rights reserved.
# *
# * Author:
# *       Appana Durga Kedareswara rao <appana.durga.rao@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import sys
import os
import re
import lopper_lib

sys.path.append(os.path.dirname(__file__))
from baremetalconfig_xlnx import scan_reg_size, get_cpu_node
from common_utils import to_cmakelist
from domain_access import domain_get_subnodes

def is_compat( node, compat_string_to_test ):
    if re.search( "module,baremetallinker_xlnx", compat_string_to_test):
        return xlnx_generate_bm_linker
    return ""


# tgt_node: is the baremetal config top level domain node number
# sdt: is the system device-tree
# options: baremetal application source path
def get_memranges(tgt_node, sdt, options):
    root_node = sdt.tree[tgt_node]
    root_sub_nodes = root_node.subnodes()
    mem_nodes = []

    domain_nodes = domain_get_subnodes(sdt.tree)
    # Get the Baremetal Domain
    for domain_node in domain_nodes:
        bm_domain = []
        if domain_node.propval('os,type') != ['']:
            if re.search('baremetal', domain_node.propval('os,type', list)[0]):
                    bm_domain.append(domain_node)
        elif domain_node.propval('cpus') != ['']:
                print('ERROR: os,type property is missing in the domain', domain_node.name)

        match_cpunode = get_cpu_node(sdt, options)

        # Check whether domain is mapped for current processor or not
        if bm_domain:
            shared_mem = []
            if not domain_node['cpus'][0] == match_cpunode.parent.phandle:
                continue
            else:
               if domain_node.propval('include') != ['']:
                   rsc_domain = lopper_lib.includes( sdt.tree, domain_node['include'] )
                   for rscdomain in rsc_domain:
                       if rscdomain.propval('memory') != ['']:
                           shared_mem.extend(rscdomain.propval('memory'))

            mem_nodes = sdt.tree.nodes('/memory@.*')
            for node in mem_nodes:
                update_shared_mem = []
                reg, size = scan_reg_size(node, node['reg'].value, 0)
                if shared_mem:
                    start_addr = shared_mem[0]
                    if start_addr in range(reg, reg+size):
                        update_shared_mem.extend(shared_mem)

                if domain_node.propval('memory') != ['']:
                    start_addr = domain_node['memory'].value[0]
                    if start_addr in range(reg, reg+size):
                        if update_shared_mem:
                            update_shared_mem.extend(domain_node['memory'].value)
                            modify_val = update_mem_node(node, update_shared_mem)
                        else:
                            modify_val = update_mem_node(node, domain_node['memory'].value)
                        node['reg'].value = modify_val
                else:
                    if update_shared_mem:
                        modify_val = update_mem_node(node, update_shared_mem)
                        node['reg'].value = modify_val

    #Maintain a static memory IP list this is needed inorder to capture proper ip name in the linker script
    xlnx_memipname = {"axi_bram": 0, "ps7_ddr": 0, "psu_ddr": 0, "psv_ddr": 0, "mig": 0, "lmb_bram": 0, "axi_noc": 0, "psu_ocm": 0,  "psv_ocm": 0, "ddr4": 0}
    for node in root_sub_nodes:
        try:
            device_type = node["device_type"].value
            if "memory" in device_type:
                mem_nodes.append(node)
        except:
           pass

    versal_noc_ch_ranges =  {"DDR_CH_1": "0x50000000000", "DDR_CH_2": "0x60000000000", "DDR_CH_3": "0x70000000000"}

    # Yocto Machine to CPU compat mapping
    match_cpunode = get_cpu_node(sdt, options)
    if not match_cpunode:
        return
    address_map = match_cpunode.parent["address-map"].value
    all_phandles = []
    ns = match_cpunode.parent["#ranges-size-cells"].value[0]
    na = match_cpunode.parent["#ranges-address-cells"].value[0]
    cells = na + ns
    tmp = na
    while tmp < len(address_map):
        all_phandles.append(address_map[tmp])
        tmp = tmp + cells + na + 1

    mem_ranges = {}
    # Remove Duplicate memory node referenecs
    mem_nodes = list(dict.fromkeys(mem_nodes))
    for node in mem_nodes:
        # Check whether the memory node is mapped to cpu cluster or not
        mem_phandles = [handle for handle in all_phandles if handle == node.phandle]
        addr_list = []
        size_list = []
        if mem_phandles:
           # Remove Duplicate phandle referenecs
           mem_phandles = list(dict.fromkeys(mem_phandles))
           indx_list = [index for index,handle in enumerate(address_map) for val in mem_phandles if handle == val]
           for inx in indx_list:
               start = [address_map[inx+i+1] for i in range(na)]
               size_list.append(address_map[inx+2*na])
               if na == 2 and start[0] != 0:
                   val = str(start[1])
                   pad = 8 - len(val)
                   val = val.ljust(pad + len(val), '0')
                   reg = int((str(hex(start[0])) + val), base=16)
                   addr_list.append(reg)
               elif na == 2:
                   addr_list.append(start[1])
               else:
                   addr_list.append(start[0])

        nac = node.parent["#address-cells"].value[0]
        nsc = node.parent["#size-cells"].value[0]
        val = node['reg'].value
        total_nodes = int(len(val)/(nac+nsc))
        name_list = [name.replace("_", "-") for name in list(xlnx_memipname.keys())]
        try:
            compat = node['compatible'].value[0]
            match = [mem for mem in name_list if mem in compat]
            for i in range(total_nodes):
                reg, size = scan_reg_size(node, val, i)
                valid_range = [addr for addr in addr_list if reg == addr or addr in range(reg, size-reg)]
                if not valid_range:
                    valid_range = [reg for index, addr in enumerate(addr_list) if reg in range(addr, size_list[index]-addr)]
                if valid_range:
                    key = match[0].replace("-", "_")
                    is_valid_noc_ch = 0
                    if "axi_noc" in key:
                        for ch_name, ran in sorted(versal_noc_ch_ranges.items(), reverse=True):
                            if int(ran, base=16) <= int(hex(valid_range[0]), base=16):
                                is_valid_noc_ch = ch_name
                                break

                    if is_valid_noc_ch:
                        linker_secname = key + str("_") + is_valid_noc_ch
                    else:
                        linker_secname = key + str("_") + str(xlnx_memipname[key])
                        xlnx_memipname[key] += 1
                    mem_ranges.update({linker_secname: [valid_range[0], size]})
        except KeyError:
            pass

    return mem_ranges

# tgt_node: is the baremetal config top level domain node number
# sdt: is the system device-tree
# options: baremetal application source path
def xlnx_generate_bm_linker(tgt_node, sdt, options):
    mem_ranges = get_memranges(tgt_node, sdt, options)
    default_ddr = None
    memtest_config = None
    machine = options['args'][0]

    try:
        memtest_config = options['args'][2]
    except IndexError:
        pass

    src_dir = options['args'][1].rstrip(os.path.sep)
    appname = os.path.basename(os.path.dirname(src_dir))
    cmake_file = os.path.join(sdt.outdir, f"{appname.capitalize()}Example.cmake")
    cfd = open(cmake_file, 'a')
    if memtest_config:
        traverse = False
    else:
        traverse = True

    mem_sec = ""
    for key, value in sorted(mem_ranges.items(), key=lambda e: e[1][1], reverse=traverse):
        if default_ddr is None:
            default_ddr = key
        start,size = value[0], value[1]
        """
        LMB BRAM initial 80 bytes being used by the linker vectors section
        Adjust the size and start address accordingly.
        """
        if "lmb_bram" in key:
            start = 80
            size -= start
        """
        PS7 DDR initial 1MB is reserved memory
        Adjust the size and start address accordingly.
        """
        if "ps7_ddr" in key:
            start = 1048576
            size -= start
        """
        For R5 PSU DDR initial 1MB is reserved for tcm
        Adjust the size and start address accordingly.
        """
        if "psu_ddr" in key and machine == "cortexr5-zynqmp" and start == 0:
            start = 1048576
            size -= start
        if "axi_noc" in key and machine == "cortexr5-versal" and start == 0:
            start = 1048576
            size -= start
        mem_sec += f'\n\t{key} : ORIGIN = {hex(start)}, LENGTH = {hex(size)}'

    ## To inline with existing tools point default ddr for linker to lower DDR
    lower_ddrs = ["axi_noc_0", "psu_ddr_0", "ps7_ddr_0"]
    has_ddr = [x for x in mem_ranges.keys() for ddr in lower_ddrs if re.search(ddr, x)]
    if has_ddr and not memtest_config:
        default_ddr = has_ddr[0]
    ## For memory tests configuration default memory should be ocm if available
    has_ocm = [x for x in mem_ranges.keys() if "ocm" in x]
    if has_ocm and memtest_config:
        default_ddr = has_ocm[0]

    cfd.write("set(DDR %s)\n" % default_ddr)
    memip_list = []
    for key, value in sorted(mem_ranges.items(), key=lambda e: e[1][1], reverse=traverse):
        memip_list.append(key)
        cfd.write("set(%s %s)\n" % (key, to_cmakelist([hex(value[0]), hex(value[1])])))
    if has_ocm and memtest_config:
        memip_list.insert(0, memip_list.pop(memip_list.index(has_ocm[0])))
    cfd.write("set(TOTAL_MEM_CONTROLLERS %s)\n" % to_cmakelist(memip_list))
    cfd.write(f'set(MEMORY_SECTION "MEMORY\n{{{mem_sec}\n}}")\n')
    return True



#/*
# * Copyright (c) 2020 Xilinx Inc. All rights reserved.
# *
# * Author:
# *       Appana Durga Kedareswara rao <appana.durga.rao@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import sys
import os
import re
import glob

sys.path.append(os.path.dirname(__file__))
import common_utils as utils
import baremetalconfig_xlnx as bm_config

def generate_drvcmake_metadata(sdt, node_list, src_dir, options):
    driver_compatlist = []
    drvname = utils.get_base_name(utils.get_dir_path(src_dir))
    yaml_file = os.path.join(utils.get_dir_path(src_dir), "data", f"{drvname}.yaml")
    if not utils.is_file(yaml_file):
        print(f"{drvname} Driver doesn't have yaml file")
        return False

    # Get the example_schema
    schema = utils.load_yaml(yaml_file)
    driver_compatlist = bm_config.compat_list(schema)
    example_schema = schema.get('examples',{})
       
    driver_nodes = []
    for compat in driver_compatlist:
        for node in node_list:
           compatlist = node['compatible'].value
           for compat_string in compatlist:
               if compat in compat_string:
                   driver_nodes.append(node)

    driver_nodes = bm_config.get_mapped_nodes(sdt, driver_nodes, options)
    nodename_list = []
    reg_list = []
    example_dict = {}
    depreg_dict = {}

    for node in driver_nodes:
        depreg_list = []
        reg, size = bm_config.scan_reg_size(node, node['reg'].value, 0)
        nodename_list.append(node.name)
        reg_list.append(hex(reg))
        
        validex_list = []
        for example,prop in example_schema.items():
            valid_ex = 0
            match_list = []
            for p in prop:
                if isinstance(p, dict):
                    for e,prop_val in p.items():
                        valid_phandle = 0
                        try:
                            val = node[e].value
                            if '' in val:
                                val = 1
                            if e == "axistream-connected":
                                reg = bm_config.get_phandle_regprop(sdt, e, val)
                                val = reg & 0xF
                            if prop_val == "phandle":
                                depreg_list.append(hex(bm_config.get_phandle_regprop(sdt, e, val)))
                                valid_phandle = 1
                        except KeyError:
                            val = 0
                        if prop_val == val:
                            match_list.append(True)
                        elif prop_val == "phandle" and valid_phandle == 1:
                            match_list.append(True)
                        elif isinstance(val, list):
                            if prop_val == val[0]:
                                match_list.append(True)
                            else:
                                match_list.append(False)
                        else:
                            match_list.append(False)
                else:
                    try:
                        valid_ex = node[p].value
                        if valid_ex:
                            match_list.append(True)
                    except KeyError:
                        match_list.append(False)

            #If all the example required conditions met it is valid example
            if False in match_list:
                valid_ex = 0
            else:
                valid_ex = 1

            if valid_ex:
                validex_list.append(example)

        example_dict.update({node.name:validex_list})
        depreg_dict.update({node.name:depreg_list})

    cmake_file = os.path.join(sdt.outdir, f"{drvname.capitalize()}Example.cmake")
    with open(cmake_file, 'a') as fd:
        fd.write(f"set(NUM_DRIVER_INSTANCES {utils.to_cmakelist(nodename_list)})\n")
        fd.write(f"set(REG_LIST {utils.to_cmakelist(reg_list)})\n")
        for index,name in enumerate(nodename_list):
            fd.write(f"set(EXAMPLE_LIST{index} {utils.to_cmakelist(example_dict[name])})\n")
            fd.write(f"set(DEPDRV_REG_LIST{index} {utils.to_cmakelist(depreg_dict[name])})\n")
            fd.write(f"list(APPEND TOTAL_EXAMPLE_LIST EXAMPLE_LIST{index})\n")
            fd.write(f"list(APPEND TOTAL_DEPDRV_REG_LIST DEPDRV_REG_LIST{index})\n")

def getmatch_nodes(sdt, node_list, yaml_file, options):
    # Get the example_schema
    schema = utils.load_yaml(yaml_file)
    driver_nodes = []
    driver_compatlist = bm_config.compat_list(schema)
    for compat in driver_compatlist:
        for node in node_list:
           compat_string = node['compatible'].value[0]
           if compat in compat_string:
               driver_nodes.append(node)

    driver_nodes = bm_config.get_mapped_nodes(sdt, driver_nodes, options)
    return driver_nodes

def getxlnx_phytype(sdt, value):
    child_node = [node for node in sdt.tree['/'].subnodes() if node.phandle == value[0]]
    phy_type = child_node[0]['xlnx,phy-type'].value[0]
    return hex(phy_type)

def lwip_topolgy(outdir, config):
    topology_fd = open(os.path.join(outdir, 'xtopology_g.c'), 'w')
    topology_str = f'''
#include "netif/xtopology.h"
#include "xil_types.h"

struct xtopology_t xtopology[] = {{'''
    for key, value in config.items():
        topology_str += f'''
    {{
        {key},
        {value},
    }},'''
    topology_str += f'''
    {{
        NULL
    }}
}};'''
    topology_fd.write(topology_str)

def generate_hwtocmake_medata(sdt, node_list, src_path, repo_path_data, options, chosen_node, symbol_node):
    src_path = src_path.rstrip(os.path.sep)
    name = utils.get_base_name(utils.get_dir_path(src_path))
    yaml_file = os.path.join(utils.get_dir_path(src_path), "data", f"{name}.yaml")

    if not utils.is_file(yaml_file):
        print(f"{drvname} Driver doesn't have yaml file")
        return False

    schema = utils.load_yaml(yaml_file)
    meta_dict = schema.get('depends',{})
    comp_type = schema.get('type',{})

    lwip = re.search("lwip", name)
    standalone = re.search("standalone", name)
    cmake_file = os.path.join(sdt.outdir, f"{name.capitalize()}Example.cmake")
    topology_data = {}
    with open(cmake_file, "a") as fd:
        lwiptype_index = 0
        for drv, prop_list in sorted(meta_dict.items(), key=lambda kv:(kv[0], kv[1])):
            if utils.is_file(repo_path_data):
                repo_schema = utils.load_yaml(repo_path_data)
                drv_data = repo_schema['driver']
                drv_dir = drv_data.get(drv,{}).get('vless','')
            else:
                drv_dir = os.path.join(repo_path_data, "XilinxProcessorIPLib", "drivers", drv)

            drv_yamlpath = os.path.join(drv_dir, "data", f"{drv}.yaml")

            if not utils.is_file(drv_yamlpath):
                print(f"{drv} yaml file {drv_yamlpath} doesnt exist")
                return False

            nodes = getmatch_nodes(sdt, node_list, drv_yamlpath, options)
            name_list = []
            for node in nodes:
                if node.propval('xlnx,name') != ['']:
                    name_list.append(node.propval('xlnx,name', list)[0])
                else:
                    name_list.append(bm_config.get_label(sdt, symbol_node, node))

            fd.write(f"set({drv.upper()}_NUM_DRIVER_INSTANCES {utils.to_cmakelist(name_list)})\n")
            for index,node in enumerate(nodes):
                val_list = []
                for prop in prop_list:
                    if prop == "reg":
                       reg,size = bm_config.scan_reg_size(node, node[prop].value, 0)
                       val = hex(reg)
                       if lwip and comp_type == "library":
                           topology_data[val] = lwiptype_index
                    elif prop == "interrupts":
                       val = bm_config.get_interrupt_prop(sdt, node, node[prop].value)
                       val = val[0]
                    elif prop == "axistream-connected":
                       val = hex(bm_config.get_phandle_regprop(sdt, prop, node[prop].value))
                    elif prop == "phy-handle":
                       try:
                           val = getxlnx_phytype(sdt, node[prop].value)
                       except KeyError:
                           val = hex(0)
                    else:
                        val = hex(node[prop].value[0])
                    val_list.append(val)
                fd.write(f"set({drv.upper()}{index}_PROP_LIST {utils.to_cmakelist(val_list)})\n")
                fd.write(f"list(APPEND TOTAL_{drv.upper()}_PROP_LIST {drv.upper()}{index}_PROP_LIST)\n")
            lwiptype_index += 1
        if standalone:
            stdin_node = bm_config.get_stdin(sdt, chosen_node, node_list)
            print(stdin_node.propval('xlnx,name'))
            if stdin_node.propval('xlnx,name') != ['']:
                fd.write(f'set(STDIN_INSTANCE "{stdin_node.propval("xlnx,name")[0]}")\n')
            else:
                fd.write(f'set(STDIN_INSTANCE "{bm_config.get_label(sdt, symbol_node, stdin_node)}")\n')

    if topology_data:
        lwip_topolgy(sdt.outdir, topology_data)

def is_compat( node, compat_string_to_test ):
    if re.search( "module,bmcmake_metadata_xlnx", compat_string_to_test):
        return xlnx_generate_cmake_metadata
    return ""


def xlnx_generate_cmake_metadata(tgt_node, sdt, options):
    root_node = sdt.tree[tgt_node]
    root_sub_nodes = root_node.subnodes()

    node_list = []
    chosen_node = ""
    symbol_node = ""
    # Traverse the tree and find the nodes having status=ok property
    for node in root_sub_nodes:
        try:
            if node.name == "chosen":
                chosen_node = node
            if node.name == "__symbols__":
                symbol_node = node
            status = node["status"].value
            if "okay" in status:
                node_list.append(node)
        except:
           pass

    src_path = options['args'][1]
    command = options['args'][2]
    repo_path = ""
    try:
        repo_path = options['args'][3]
    except IndexError:
        pass

    if command == "drvcmake_metadata":
        generate_drvcmake_metadata(sdt, node_list, src_path, options)
    elif command == "hwcmake_metadata":
        generate_hwtocmake_medata(sdt, node_list, src_path, repo_path, options, chosen_node, symbol_node)
    return True



#/*
# * Copyright (c) 2019,2020 Xilinx Inc. All rights reserved.
# *
# * Author:
# *       Bruce Ashfield <bruce.ashfield@xilinx.com>
# *
# * SPDX-License-Identifier: BSD-3-Clause
# */

import copy
import struct
import sys
import types
import unittest
import os
import getopt
import re
import subprocess
import shutil
from pathlib import Path
from pathlib import PurePath
from io import StringIO
import contextlib
import importlib
from lopper import Lopper
from lopper import LopperFmt
import lopper
from lopper.tree import *
from re import *
from enum import Enum

sys.path.append(os.path.dirname(__file__))
from openamp_xlnx_common import *

RPU_PATH = "/rpu@ff9a0000"
REMOTEPROC_D_TO_D = "openamp,remoteproc-v1"
RPMSG_D_TO_D = "openamp,rpmsg-v1"
output_file = "openamp-channel-info.txt"

class CPU_CONFIG(Enum):
    RPU_LOCKSTEP = 0
    RPU_SPLIT = 1

class RPU_CORE(Enum):
    RPU_0 = 0
    RPU_1 = 1

# This is used for YAML representation
# after this is parsed, the above enums are used for internal record keeping.
class CLUSTER_CONFIG(Enum):
    RPU_LOCKSTEP = 0
    RPU_0 = 1
    RPU_1 = 2

def is_compat( node, compat_string_to_test ):
    if re.search( "openamp,xlnx-rpu", compat_string_to_test):
        return xlnx_openamp_rpu
    return ""

def xlnx_rpmsg_native_update_carveouts(tree, elfload_node,
                                       native_shm_mem_area_start, native_shm_mem_area_size,
                                       native_amba_shm_node):
    # start and size of reserved-mem rproc0
    start = elfload_node.props("start")[0].value
    size = elfload_node.props("size")[0].value

    # update size if applicable
    if start < native_shm_mem_area_start:
        native_shm_mem_area_start = start
    native_shm_mem_area_size += size

    # update rproc0 size
    elfload_res_mem_node = tree["/reserved-memory/" + elfload_node.name]
    elfload_res_mem_reg = elfload_res_mem_node.props("reg")[0].value
    elfload_res_mem_reg[3] = native_shm_mem_area_size
    elfload_res_mem_node.props("reg")[0].value = elfload_res_mem_reg

    # FIXME: update demos so that start of shm is not hard-coded offset
    native_shm_mem_area_start += 0x20000
    shm_space = [0, native_shm_mem_area_start, 0, native_shm_mem_area_size]
    native_amba_shm_node + LopperProp(name="reg", value=shm_space)

    return True


native_shm_node_count = 0
def xlnx_rpmsg_construct_carveouts(tree, carveouts, rpmsg_carveouts, native,
                                   amba_node = None, elfload_node = None, verbose = 0 ):
    global native_shm_node_count
    res_mem_node = tree["/reserved-memory"]
    native_amba_shm_node = None
    native_shm_mem_area_size = 0
    native_shm_mem_area_start = 0xFFFFFFFF

    if amba_node != None:
        native_amba_shm_node = LopperNode(-1, amba_node.abs_path + "/shm" + str(native_shm_node_count))
        native_amba_shm_node + LopperProp(name="compatible", value="shm_uio")
        tree.add(native_amba_shm_node)
        tree.resolve()
        native_shm_node_count += 1


    # only applicable for DDR carveouts
    for carveout in carveouts:
        # SRAM banks have status prop
        # SRAM banks are not in reserved memory
        # FIXME  skip SRAM banks in RPMsg support
        if carveout.props("status") != []:
            continue
        elif carveout.props("no-map") != []:
            start = carveout.props("start")[0].value
            size = carveout.props("size")[0].value
            # handle native RPMsg
            if native:
                if start < native_shm_mem_area_start:
                    native_shm_mem_area_start = start
                native_shm_mem_area_size += size
                # add more space for shared buffers
                if 'vdev0buffer' in carveout.name:
                    native_shm_mem_area_size += size
            else:

                new_node =  LopperNode(-1, "/reserved-memory/"+carveout.name)
                new_node + LopperProp(name="no-map")
                new_node + LopperProp(name="reg", value=[0, start, 0, size])

                if "vdev0buffer" in carveout.name:
                    new_node + LopperProp(name="compatible", value="shared-dma-pool")

                tree.add(new_node)
                tree.resolve()

                if new_node.phandle == 0:
                    new_node.phandle = new_node.phandle_or_create()

                new_node + LopperProp(name="phandle", value = new_node.phandle)
                rpmsg_carveouts.append(new_node)

        else:
            print("WARNING: invalid remoteproc elfload carveout", carveout)
            return False


    if native:
        ret = xlnx_rpmsg_native_update_carveouts(tree, elfload_node,
                                                 native_shm_mem_area_start, native_shm_mem_area_size,
                                                 native_amba_shm_node)
        if not ret:
            return ret

    return True


def xlnx_rpmsg_ipi_parse(tree, node, openamp_channel_info,
                         remote_node, channel_id, native,
                         verbose = 0 ):
    remote = node.props("remote")
    carveouts_prop = node.props("carveouts")
    amba_node = None
    ipi_id_prop_name = "xlnx,ipi-id"
    host_to_remote_ipi = None
    remote_to_host_ipi = None

    # collect host ipi
    host_ipi_prop = node.props("mbox")
    if host_ipi_prop == []:
        print("WARNING: ", node, " is missing mbox property")
        return False

    host_ipi_prop = host_ipi_prop[0]
    host_ipi = tree.pnode(host_ipi_prop.value[0])

    # collect corresponding remote rpmsg relation
    remote_node = tree.pnode( remote[0].value[0] )
    remote_rpmsg_relation = None
    try:
        remote_rpmsg_relation = tree[remote_node.abs_path + "/domain-to-domain/rpmsg-relation"]
    except:
        print("WARNING: ", remote_node, " is missing rpmsg relation")
        return False

    # collect remote ipi
    remote_ipi_prop = remote_rpmsg_relation.props("mbox")
    if remote_ipi_prop == []:
        print("WARNING: ", remote_node, " is missing mbox property")
        return False

    remote_ipi_prop = remote_ipi_prop[0]
    remote_ipi = tree.pnode(remote_ipi_prop.value[0])

    host_ipi_id = host_ipi.props(ipi_id_prop_name)
    if host_ipi_id == []:
        print("WARNING", host_ipi, " does not have property name: ", ipi_id_prop_name)
    host_ipi_id = host_ipi_id[0]

    remote_ipi_id = remote_ipi.props(ipi_id_prop_name)
    if remote_ipi_id == []:
        print("WARNING", remote_ipi, " does not have property name: ", ipi_id_prop_name)
        return False
    remote_ipi_id = remote_ipi_id[0]

    # find host to remote buffers
    host_to_remote_ipi_channel = None
    for subnode in host_ipi.subnodes():
        subnode_ipi_id = subnode.props(ipi_id_prop_name)
        if subnode_ipi_id != [] and remote_ipi_id.value[0] == subnode_ipi_id[0].value[0]:
            host_to_remote_ipi_channel = subnode
    if host_to_remote_ipi_channel == None:
        print("WARNING no host to remote IPI channel has been found.")
        return False

    # find remote to host buffers
    remote_to_host_ipi_channel = None
    for subnode in remote_ipi.subnodes():
        subnode_ipi_id = subnode.props(ipi_id_prop_name)
        if subnode_ipi_id != [] and host_ipi_id.value[0] == subnode_ipi_id[0].value[0]:
            remote_to_host_ipi_channel = subnode
    if remote_to_host_ipi_channel == None:
        print("WARNING no remote to host IPI channel has been found.")
        return False

    # set platform
    platform = None
    root_node = tree["/"]
    root_compat = root_node.props("compatible")[0].value
    for compat in root_compat:
        if "zynqmp" in compat:
            platform = SOC_TYPE.ZYNQMP
            break
        elif "versal" in compat:
            platform = SOC_TYPE.VERSAL
            break
    openamp_channel_info["platform"] = platform

    # store IPI IRQ Vector IDs
    platform =  openamp_channel_info["platform"]
    host_ipi_base = host_ipi.props("reg")[0][1]
    remote_ipi_base = remote_ipi.props("reg")[0][1]

    if SOC_TYPE.ZYNQMP == platform:
        if host_ipi_base not in zynqmp_ipi_to_irq_vect_id.keys():
            print("WARNING: host IPI", hex(host_ipi_base), "not in IRQ VECTOR ID Mapping for ZU+")
            return False
        if remote_ipi_base not in zynqmp_ipi_to_irq_vect_id.keys():
            print("WARNING: remote IPI", hex(remote_ipi_base), "not in IRQ VECTOR ID Mapping for ZU+")
            return False

        openamp_channel_info["host_ipi_base"+channel_id] = host_ipi_base
        openamp_channel_info["host_ipi_irq_vect_id"+channel_id] = zynqmp_ipi_to_irq_vect_id[host_ipi_base]
        openamp_channel_info["remote_ipi_base"+channel_id] = remote_ipi_base
        openamp_channel_info["remote_ipi_irq_vect_id"+channel_id] = zynqmp_ipi_to_irq_vect_id[remote_ipi_base]
    elif SOC_TYPE.VERSAL == platform:
        if host_ipi_base not in versal_ipi_to_irq_vect_id.keys():
            print("WARNING: host IPI", hex(host_ipi_base), "not in IRQ VECTOR ID Mapping for ZU+")
            return False
        if remote_ipi_base not in versal_ipi_to_irq_vect_id.keys():
            print("WARNING: remote IPI", hex(remote_ipi_base), "not in IRQ VECTOR ID Mapping for ZU+")
            return False

        openamp_channel_info["host_ipi_base"+channel_id] = host_ipi_base
        openamp_channel_info["host_ipi_irq_vect_id"+channel_id] = versal_ipi_to_irq_vect_id[host_ipi_base]
        openamp_channel_info["remote_ipi_base"+channel_id] = remote_ipi_base
        openamp_channel_info["remote_ipi_irq_vect_id"+channel_id] = versal_ipi_to_irq_vect_id[remote_ipi_base]
    else:
        print("Unsupported platform")

    openamp_channel_info["host_ipi_"+channel_id] = host_ipi
    openamp_channel_info["remote_ipi_"+channel_id] = remote_ipi
    openamp_channel_info["rpmsg_native_"+channel_id] = native
    openamp_channel_info["host_to_remote_ipi_channel_" + channel_id] = host_to_remote_ipi_channel
    openamp_channel_info["remote_to_host_ipi_channel_" + channel_id] = remote_to_host_ipi_channel

    return True


def xlnx_rpmsg_setup_host_controller(tree, controller_parent, gic_node_phandle,
                                     host_ipi):
    controller_parent + LopperProp(name="compatible",value="xlnx,zynqmp-ipi-mailbox")
    controller_parent + LopperProp(name="interrupt-parent", value = [gic_node_phandle])
    controller_parent + LopperProp(name="#address-cells",value=1)
    controller_parent + LopperProp(name="#size-cells",value=1)
    controller_parent + LopperProp(name="ranges")

    host_interrupts_prop = host_ipi.props("interrupts")
    if host_interrupts_prop == []:
        print("WARNING: host ipi ", host_ipi, " missing interrupts property")
        return False

    host_interrupts_pval = host_interrupts_prop[0].value
    controller_parent + LopperProp(name="interrupts", value = copy.deepcopy(host_interrupts_pval))
    controller_parent + LopperProp(name="xlnx,ipi-id", value = host_ipi.props("xlnx,ipi-id")[0].value)

    tree.add(controller_parent)
    tree.resolve()

    if controller_parent.phandle == 0:
        controller_parent.phandle = controller_parent.phandle_or_create()

    controller_parent + LopperProp(name="phandle", value = controller_parent.phandle)

    return True


controller_parent_index = 1
def xlnx_rpmsg_create_host_controller(tree, host_ipi, gic_node_phandle, verbose = 0):
    global controller_parent_index
    controller_parent = None
    ctr_parent_name = "/zynqmp_ipi" + str(controller_parent_index)
    setup_host_controller = True
    try:
        controller_parent = tree[ctr_parent_name]

        ctr_parent_ipi_id = controller_parent.props("xlnx,ipi-id")
        if ctr_parent_ipi_id != [] and ctr_parent_ipi_id[0].value[0] == host_ipi.props("xlnx,ipi-id")[0].value[0]:
            setup_host_controller = False
        else:
            setup_host_controller = True
            controller_parent_index += 1
            # make new parent controller with updated name
            ctr_parent_name = "/zynqmp_ipi" + str(controller_parent_index)
            controller_parent = LopperNode(-1, ctr_parent_name)
    except:
        controller_parent = LopperNode(-1, ctr_parent_name)

    # check if host controller needs to be setup before adding props to it
    if setup_host_controller:
        ret = xlnx_rpmsg_setup_host_controller(tree, controller_parent,
                                               gic_node_phandle, host_ipi)
        if not ret:
            return False

    return controller_parent


def xlnx_rpmsg_native_update_ipis(tree, amba_node, openamp_channel_info, gic_node_phandle,
                                  amba_ipi_node_index, host_ipi):
    amba_node = openamp_channel_info["amba_node"]
    amba_ipi_node = LopperNode(-1, amba_node.abs_path + "/openamp_ipi" + str(amba_ipi_node_index))
    amba_ipi_node + LopperProp(name="compatible",value="ipi_uio")
    amba_ipi_node + LopperProp(name="interrupts",value=copy.deepcopy(host_ipi.props("interrupts")[0].value))
    amba_ipi_node + LopperProp(name="interrupt-parent",value=[gic_node_phandle])
    reg_val = copy.deepcopy(host_ipi.props("reg")[0].value)
    reg_val[3] = 0x1000
    amba_ipi_node + LopperProp(name="reg",value=reg_val)
    tree.add(amba_ipi_node)
    tree.resolve()
    amba_ipi_node_index += 1

    # check if there is a kernelspace IPI that has same interrupts
    # if so then error out
    for n in tree["/"].subnodes():
        compat_str = n.props("compatible")
        if compat_str != [] and compat_str[0].value == "xlnx,zynqmp-ipi-mailbox" and \
            n.props("interrupts")[0].value[1] ==  host_ipi.props("interrupts")[0].value[1]:
            # check if subnode exists with the correct structure
            # before erroring out. There may be IP IPIs that this
            # catches
            for sub_n in n.subnodes():
                if sub_n.props("reg-names") != []:
                    print("WARNING:  conflicting userspace and kernelspace for host ipi: ", host_ipi)
                    return False

    return True


def xlnx_rpmsg_kernel_update_mboxes(tree, host_ipi, remote_ipi, gic_node_phandle,
                                  openamp_channel_info, channel_id,
                                  core_node, remote_controller_node, rpu_core):
    remote_controller_node + LopperProp(name="phandle", value = remote_controller_node.phandle)
    remote_controller_node + LopperProp(name="#mbox-cells", value = 1)
    remote_controller_node + LopperProp(name="xlnx,ipi-id", value = remote_ipi.props("xlnx,ipi-id")[0].value)
    reg_names_val = [ "local_request_region", "local_response_region",
                      "remote_request_region","remote_response_region"]
    remote_controller_node + LopperProp(name="reg-names", value = reg_names_val)

    cpu_config = str(openamp_channel_info["cpu_config"+channel_id].value)
    host_to_remote_ipi_channel = openamp_channel_info["host_to_remote_ipi_channel_"+channel_id]
    remote_to_host_ipi_channel= openamp_channel_info["remote_to_host_ipi_channel_"+channel_id]
    response_buf_str = "xlnx,ipi-rsp-msg-buf"
    request_buf_str = "xlnx,ipi-req-msg-buf"

    host_remote_response = host_to_remote_ipi_channel.props(response_buf_str)
    if host_remote_response == []:
        print("WARNING: host_remote_response ", host_to_remote_ipi_channel, "is missing property name: ", response_buf_str)
        return False
    host_remote_request = host_to_remote_ipi_channel.props(request_buf_str)
    if host_remote_request == []:
        print("WARNING: host_remote_request ", host_to_remote_ipi_channel, " is missing property name: ", request_buf_str)
        return False
    remote_host_response = remote_to_host_ipi_channel.props(response_buf_str)
    if remote_host_response == []:
        print("WARNING: remote_host_response ", remote_to_host_ipi_channel, " is missing property name: ", response_buf_str)
        return False
    remote_host_request = remote_to_host_ipi_channel.props(request_buf_str)
    if remote_host_request == []:
        print("WARNING: remote_host_request ", remote_to_host_ipi_channel, " is missing property name: ", request_buf_str)
        return False

    host_remote_response = host_remote_response[0].value[0]
    host_remote_request = host_remote_request[0].value[0]
    remote_host_response = remote_host_response[0].value[0]
    remote_host_request = remote_host_request[0].value[0]

    remote_controller_node + LopperProp(name="reg", value = [
                                                         host_remote_response, 0x20,
                                                         host_remote_request, 0x20,
                                                         remote_host_request, 0x20,
                                                         remote_host_response, 0x20])

    core_node + LopperProp(name="mboxes", value = [remote_controller_node.phandle, 0, remote_controller_node.phandle, 1])
    return True

def xlnx_rpmsg_kernel_update_ipis(tree, host_ipi, remote_ipi, gic_node_phandle,
                                  core_node, openamp_channel_info, channel_id):
    controller_parent = xlnx_rpmsg_create_host_controller(tree, host_ipi, gic_node_phandle)

    # check if there is a userspace IPI that has same interrupts
    # if so then error out
    for n in tree["/"].subnodes():
        compat_str = n.props("compatible")
        if compat_str != [] and compat_str[0].value == "ipi_uio" and \
            n.props("interrupts")[0].value[1] ==  host_ipi.props("interrupts")[0].value[1]:
            print("WARNING:  conflicting userspace and kernelspace for host ipi: ", host_ipi)
            return False

    if controller_parent == False:
        return False

    rpu_core = openamp_channel_info["rpu_core" + channel_id]
    remote_controller_node = LopperNode(-1, controller_parent.abs_path + "/ipi_mailbox_rpu" + str(rpu_core.value))
    tree.add(remote_controller_node)
    tree.resolve()

    if remote_controller_node.phandle == 0:
        remote_controller_node.phandle = remote_controller_node.phandle_or_create()


    return xlnx_rpmsg_kernel_update_mboxes(tree, host_ipi, remote_ipi, gic_node_phandle,
                                           openamp_channel_info, channel_id,
                                           core_node, remote_controller_node, rpu_core)


amba_ipi_node_index = 0
def xlnx_rpmsg_update_ipis(tree, host_ipi, remote_ipi, core_node, channel_id, openamp_channel_info, verbose = 0 ):
    global amba_ipi_node_index
    native = openamp_channel_info["rpmsg_native_"+ channel_id]
    platform = openamp_channel_info["platform"]
    controller_parent = None
    amba_node = None
    gic_node_phandle = None

    if platform == SOC_TYPE.VERSAL:
        gic_node_phandle = tree["/apu-bus/interrupt-controller@f9000000"].phandle
    elif platform == SOC_TYPE.ZYNQMP:
        gic_node_phandle = tree["/apu-bus/interrupt-controller@f9010000"].phandle
    else:
        print("invalid platform")
        return False

    if native:
        return xlnx_rpmsg_native_update_ipis(tree, amba_node, openamp_channel_info, gic_node_phandle,
                                             amba_ipi_node_index, host_ipi)
    else:
        return xlnx_rpmsg_kernel_update_ipis(tree, host_ipi, remote_ipi, gic_node_phandle,
                                             core_node, openamp_channel_info, channel_id)


def xlnx_rpmsg_update_tree(tree, node, channel_id, openamp_channel_info, verbose = 0 ):
    cpu_config =  openamp_channel_info["cpu_config"+channel_id]
    carveouts_nodes = openamp_channel_info["carveouts_"+ channel_id]
    host_ipi = openamp_channel_info["host_ipi_"+ channel_id]
    remote_ipi = openamp_channel_info["remote_ipi_"+ channel_id]
    native = openamp_channel_info["rpmsg_native_"+ channel_id]
    amba_node = None
    elfload_node = None
    if native:
        amba_node = openamp_channel_info["amba_node"]
    rpu_core = openamp_channel_info["rpu_core" + channel_id]
    rpmsg_carveouts = []


    # if Amba node exists, then this is for RPMsg native.
    # in this case find elfload node in case of native RPMsg as it may be contiguous
    # for AMBA Shm Node
    if native:
        for node in openamp_channel_info["elfload"+ channel_id]:
            if node.props("start") != []:
                elfload_node = node

    ret = xlnx_rpmsg_construct_carveouts(tree, carveouts_nodes, rpmsg_carveouts, native,
                                         amba_node=amba_node, elfload_node=elfload_node, verbose=verbose)
    if ret == False:
        return ret

    core_node = tree["/rf5ss@ff9a0000/r5f_" + str(rpu_core.value)]
    mem_region_prop = core_node.props("memory-region")[0]

    # add rpmsg carveouts to cluster core node if using rpmsg kernel driver
    new_mem_region_prop_val = mem_region_prop.value
    if not native:
        for rc in rpmsg_carveouts:
            new_mem_region_prop_val.append(rc.phandle)
        # update property with new values
        mem_region_prop.value = new_mem_region_prop_val

    ret = xlnx_rpmsg_update_ipis(tree, host_ipi, remote_ipi, core_node, channel_id, openamp_channel_info, verbose)
    if ret != True:
        return False

    tree - host_ipi
    tree - remote_ipi

    return True


channel_index = 0
def xlnx_construct_text_file(openamp_channel_info, channel_id, verbose = 0 ):
    global channel_index
    text_file_contents = ""
    rpmsg_native = openamp_channel_info["rpmsg_native_"+channel_id]
    carveouts = openamp_channel_info["carveouts_"+channel_id]
    elfload = openamp_channel_info["elfload"+channel_id]
    tx = None
    rx = None

    for c in carveouts:
        base = hex(c.props("start")[0].value)
        size = hex(c.props("size")[0].value)
        name = ""
        if "vring0" in c.name:
            name = "VRING0"
            tx = base
        elif "vring1" in c.name:
            name = "VRING1"
            rx = base
        else:
            name = "VDEV0BUFFER"

        text_file_contents += "CHANNEL"+str(channel_index)+name+"BASE=\""+base+"\"\n"
        text_file_contents += "CHANNEL"+str(channel_index)+name+"SIZE=\""+size+"\"\n"

    if not rpmsg_native:
        tx = "FW_RSC_U32_ADDR_ANY"
        rx = "FW_RSC_U32_ADDR_ANY"

    elfbase = None
    elfsize = None
    for e in elfload:
        if e.props("start") != []: # filter to only parse ELF LOAD node
            elfbase = hex(e.props("start")[0].value)
            elfsize = hex(e.props("size")[0].value)
            break

    host_ipi = openamp_channel_info["host_ipi_" + channel_id]
    host_ipi_bitmask = hex(host_ipi.props("xlnx,ipi-bitmask")[0].value[0])
    host_ipi_irq_vect_id = hex(openamp_channel_info["host_ipi_irq_vect_id" + channel_id])
    host_ipi_base = hex(openamp_channel_info["host_ipi_base"+channel_id])

    remote_ipi = openamp_channel_info["remote_ipi_" + channel_id]
    remote_ipi_bitmask = hex(remote_ipi.props("xlnx,ipi-bitmask")[0].value[0])
    remote_ipi_irq_vect_id = hex(openamp_channel_info["remote_ipi_irq_vect_id" + channel_id])
    remote_ipi_base = hex(openamp_channel_info["remote_ipi_base"+channel_id])

    text_file_contents += "CHANNEL"+str(channel_index)+name+"RX=\""+rx+"\"\n"
    text_file_contents += "CHANNEL"+str(channel_index)+name+"TX=\""+tx+"\"\n"
    text_file_contents += "CHANNEL"+str(channel_index)+"ELFBASE=\""+elfbase+"\"\n"
    text_file_contents += "CHANNEL"+str(channel_index)+"ELFSIZE=\""+elfsize+"\"\n"
    text_file_contents += "CHANNEL"+str(channel_index)+"TO_HOST=\""+ host_ipi_base  +"\"\n"
    text_file_contents += "CHANNEL"+str(channel_index)+"TO_HOST-BITMASK=\"" + host_ipi_bitmask + "\"\n"
    text_file_contents += "CHANNEL"+str(channel_index)+"TO_HOST-IPIIRQVECTID=\"" + host_ipi_irq_vect_id + "\"\n"
    text_file_contents += "CHANNEL"+str(channel_index)+"TO_REMOTE=\""+ remote_ipi_base  +"\"\n"
    text_file_contents += "CHANNEL"+str(channel_index)+"TO_REMOTE-BITMASK=\"" + remote_ipi_bitmask + "\"\n"
    text_file_contents += "CHANNEL"+str(channel_index)+"TO_REMOTE-IPIIRQVECTID=\"" + remote_ipi_irq_vect_id + "\"\n"

    f = open(output_file, "w")
    f.write(text_file_contents)
    f.close()

    channel_index += 1
    return True


def xlnx_rpmsg_parse(tree, node, openamp_channel_info, verbose = 0 ):
    # Xilinx OpenAMP subroutine to collect RPMsg information from RPMsg
    # relation
    remote = node.props("remote")
    carveouts_nodes = []
    carveouts_prop = node.props("carveouts")
    amba_node = None

    # skip rpmsg remote node which will link to its host via 'host' property
    if node.props("host") != []:
        return True

    if remote == []:
        print("WARNING: ", node, "is missing remote property")
        return False

    remote_node = tree.pnode( remote[0].value[0] )
    channel_id = "_"+node.parent.parent.name+"_"+remote_node.name


    if carveouts_prop == []:
        print("WARNING: ", node, " is missing elfload property")
        return False
    else:
        carveouts_prop = carveouts_prop[0].value

    for p in carveouts_prop:
        carveouts_nodes.append ( tree.pnode(p) )

    openamp_channel_info["carveouts_"+channel_id] = carveouts_nodes

    # rpmsg native?
    native = node.props("openamp-xlnx-native")
    if native != [] and native[0].value[0] == 1:
        native = True
        # if native is true, then find and store amba bus
        # to store IPI and SHM nodes
        try:
            amba_node = tree["/axi"]
        except:
            amba_node = LopperNode(-1, "/axi")
            amba_node + LopperProp(name="u-boot,dm-pre-reloc")
            amba_node + LopperProp(name="ranges")
            amba_node + LopperProp(name="#address-cells", value = 2)
            amba_node + LopperProp(name="#size-cells", value = 2)
            tree.add(amba_node)
            tree.resolve()

        openamp_channel_info["amba_node"] = amba_node
    else:
        native = False

    ret = xlnx_rpmsg_ipi_parse(tree, node, openamp_channel_info,
                         remote_node, channel_id, native, verbose)
    if ret != True:
        return False

    ret = xlnx_rpmsg_update_tree(tree, node, channel_id, openamp_channel_info, verbose )
    if ret != True:
        return False

    # generate text file
    ret = xlnx_construct_text_file(openamp_channel_info, channel_id, verbose)
    if not ret:
        return ret

    # remove definitions
    defn_node =  tree["/definitions"]
    tree - defn_node

    return True


# tests for a bit that is set, going fro 31 -> 0 from MSB to LSB
def check_bit_set(n, k):
    if n & (1 << (k)):
        return True

    return False


def determine_cpus_config(remote_domain):
  cpus_prop_val = remote_domain.propval("cpus")
  cpu_config = None # split or lockstep

  if cpus_prop_val != ['']:
    if len(cpus_prop_val) != 3:
      print("rpu cluster cpu prop invalid len")
      return -1
    for i in CLUSTER_CONFIG:
        if i.value == cpus_prop_val[1]:
            if CLUSTER_CONFIG.RPU_LOCKSTEP.value == cpus_prop_val[1]:
                return CPU_CONFIG.RPU_LOCKSTEP
            else:
                return CPU_CONFIG.RPU_SPLIT

    # if here then no match
    print("WARNING: invalid cpus for ", remote_node, cpus_prop_val)
    return -1


def determinte_rpu_core(cpu_config, remote_node, remote_prop):
    remote_cpus = remote_node.props("cpus")[0]

    if cpu_config == CPU_CONFIG.RPU_LOCKSTEP:
        return RPU_CORE.RPU_0
    elif cpu_config == CPU_CONFIG.RPU_SPLIT:
        if remote_cpus[1] == 0x1:
            return RPU_CORE.RPU_0
        elif remote_cpus[1] == 0x2:
            return RPU_CORE.RPU_1
        else:
            print("WARNING: invalid cpus for ", remote_node)
            return False
    else:
        print("WARNING: invalid cpu config: ", cpu_config)
        return False


def xlnx_remoteproc_construct_carveouts(tree, carveouts, new_ddr_nodes, verbose = 0 ):
    res_mem_node = None

    try:
        res_mem_node = tree["/reserved-memory"]
    except KeyError:
        res_mem_node = LopperNode(-1, "/reserved-memory")
        res_mem_node + LopperProp(name="#address-cells",value=2)
        res_mem_node + LopperProp(name="#size-cells",value=2)
        res_mem_node + LopperProp(name="ranges",value=[])
        tree.add(res_mem_node)

    # only applicable for DDR carveouts
    for carveout in carveouts:
        # SRAM banks have status prop
        # SRAM banks are not in reserved memory
        if carveout.props("status") != []:
            continue
        elif carveout.props("no-map") != []:
            start = carveout.props("start")[0].value
            size = carveout.props("size")[0].value
            new_node =  LopperNode(-1, "/reserved-memory/"+carveout.name)
            new_node + LopperProp(name="no-map")
            new_node + LopperProp(name="reg", value=[0, start, 0, size])
            tree.add(new_node)

            phandle_val = new_node.phandle_or_create()

            new_node + LopperProp(name="phandle", value = phandle_val)
            new_node.phandle = phandle_val
            new_ddr_nodes.append(new_node.phandle)
        else:
            print("WARNING: invalid remoteproc elfload carveout", carveout)
            return False

    return True


def xlnx_remoteproc_construct_cluster(tree, cpu_config, elfload_nodes, new_ddr_nodes, host_node, remote_node, openamp_channel_info, verbose = 0):
    channel_id = "_"+host_node.name+"_"+remote_node.name
    cluster_node = None
    if cpu_config in [ CPU_CONFIG.RPU_LOCKSTEP, CPU_CONFIG.RPU_SPLIT ]:
        try:
            cluster_node = tree["/rf5ss@ff9a0000"]
        except KeyError:
            cluster_node = LopperNode(-1, "/rf5ss@ff9a0000")
            cluster_node + LopperProp(name="compatible", value = "xlnx,zynqmp-r5-remoteproc")
            cluster_node + LopperProp(name="#address-cells", value = 0x2)
            cluster_node + LopperProp(name="#size-cells", value = 0x2)
            cluster_node + LopperProp(name="ranges", value= [])
            cluster_node + LopperProp(name="xlnx,cluster-mode", value = cpu_config.value)
            cluster_node + LopperProp(name="reg", value = [0, 0xff9a0000, 0, 0x10000])
            tree.add(cluster_node)

        rpu_core = openamp_channel_info["rpu_core"+channel_id]
        rpu_core_pd_prop = openamp_channel_info["rpu_core_pd_prop"+channel_id]

        core_node = LopperNode(-1, "/rf5ss@ff9a0000/r5f_" + str(rpu_core.value))
        core_node + LopperProp(name="compatible", value = "xilinx,r5f")
        core_node + LopperProp(name="#address-cells", value = 2)
        core_node + LopperProp(name="#size-cells", value = 2)
        core_node + LopperProp(name="ranges", value=[])
        core_node + LopperProp(name="power-domain", value=copy.deepcopy(rpu_core_pd_prop.value))
        core_node + LopperProp(name="mbox-names", value = ["tx", "rx"]);

        tree.add(core_node)

        srams = []
        memory_region = []
        for carveout in elfload_nodes:
            if carveout.props("status") != []:
                srams.append(carveout.phandle)
                # FIXME for each sram, add 'power-domain' prop for kernel driver
                carveout + LopperProp(name="power-domain",
                                      value=copy.deepcopy( carveout.props("power-domains")[0].value ))

        # there may be new nodes created in linux reserved-memory node to account for
        for phandle_val in new_ddr_nodes:
            memory_region.append(phandle_val)

        core_node + LopperProp(name="memory-region", value=memory_region)
        core_node + LopperProp(name="sram", value=srams)

    else:
        return False

    return True

def xlnx_remoteproc_update_tree(tree, node, remote_node, openamp_channel_info, verbose = 0 ):
    channel_id = "_"+node.parent.parent.name+"_"+remote_node.name
    cpu_config =  openamp_channel_info["cpu_config"+channel_id]
    elfload_nodes = openamp_channel_info["elfload"+channel_id]
    new_ddr_nodes = []

    ret = xlnx_remoteproc_construct_carveouts(tree, elfload_nodes, new_ddr_nodes, verbose)
    if ret == False:
        return ret
    ret = xlnx_remoteproc_construct_cluster(tree, cpu_config, elfload_nodes, new_ddr_nodes, node.parent.parent, remote_node, openamp_channel_info, verbose = 0)
    if ret == False:
        return ret

    return True


def xlnx_remoteproc_parse(tree, node, openamp_channel_info, verbose = 0 ):
    # Xilinx OpenAMP subroutine to collect RPMsg information from Remoteproc
    # relation
    remote = node.props("remote")
    elfload_nodes = []
    elfload_prop = node.props("elfload")

    if remote == []:
        print("WARNING: ", node, "is missing remote property")
        return False

    remote_node = tree.pnode( remote[0].value[0] )
    cpu_config = determine_cpus_config(remote_node)

    if cpu_config in [ CPU_CONFIG.RPU_LOCKSTEP, CPU_CONFIG.RPU_SPLIT]:
        rpu_core = determinte_rpu_core(cpu_config, remote_node, remote[0] )
        if not rpu_core:
            return False
    else:
        return False

    if elfload_prop == []:
        print("WARNING: ", node, " is missing elfload property")
        return False
    else:
        elfload_prop = elfload_prop[0].value

    for p in elfload_prop:
        elfload_nodes.append ( tree.pnode(p) )

    rpu_cluster_node = tree.pnode(remote_node.props("cpus")[0].value[0])
    rpu_core_node = rpu_cluster_node.abs_path + "/cpu@"

    if CPU_CONFIG.RPU_SPLIT == cpu_config and RPU_CORE.RPU_1 == rpu_core:
        rpu_core_node = tree[rpu_core_node+"1"]
    elif CPU_CONFIG.RPU_LOCKSTEP == cpu_config:
        rpu_core_node = tree[rpu_core_node+"0"]
    elif CPU_CONFIG.RPU_SPLIT == cpu_config and RPU_CORE.RPU_0 == rpu_core:
        rpu_core_node = tree[rpu_core_node+"0"]
    else:
        print("WARNING invalid RPU and CPU config for relation. cpu: ", cpu_config, " rpu: ", rpu_core)
        return False

    if rpu_core_node.props("power-domains") == []:
        print("WARNING: RPU core does not have power-domains property.")
        return False

    rpu_core_pd_prop = rpu_core_node.props("power-domains")[0]

    channel_id = "_"+node.parent.parent.name+"_"+remote_node.name
    openamp_channel_info["elfload"+channel_id] = elfload_nodes
    openamp_channel_info["rpu_core_pd_prop"+channel_id] = rpu_core_pd_prop
    openamp_channel_info["cpu_config"+channel_id] = cpu_config
    openamp_channel_info["rpu_core"+channel_id] = rpu_core

    return xlnx_remoteproc_update_tree(tree, node, remote_node, openamp_channel_info, verbose = 0 )


def xlnx_openamp_parse(sdt, verbose = 0 ):
    # Xilinx OpenAMP subroutine to parse OpenAMP Channel
    # information and generate Device Tree information.
    tree = sdt.tree
    ret = -1
    openamp_channel_info = {}

    for n in tree["/domains"].subnodes():
            node_compat = n.props("compatible")
            if node_compat != []:
                node_compat = node_compat[0].value

                if node_compat == REMOTEPROC_D_TO_D:
                    ret = xlnx_remoteproc_parse(tree, n, openamp_channel_info, verbose)
                elif node_compat == RPMSG_D_TO_D:
                    ret = xlnx_rpmsg_parse(tree, n, openamp_channel_info, verbose)

                if ret == False:
                    return ret

    return True

def xlnx_openamp_rpmsg_expand(tree, subnode, verbose = 0 ):
    # Xilinx-specific YAML expansion of RPMsg description.
    ret = resolve_host_remote( tree, subnode, verbose)
    if ret == False:
        return ret
    ret = resolve_rpmsg_mbox( tree, subnode, verbose)
    if ret == False:
        return ret
    ret = resolve_rpmsg_carveouts( tree, subnode, verbose)
    if ret == False:
        return ret


    return True


def xlnx_openamp_remoteproc_expand(tree, subnode, verbose = 0 ):
    # Xilinx-specific YAML expansion of Remoteproc description.
    ret = resolve_host_remote( tree, subnode, verbose)
    if ret == False:
        return ret
    ret = resolve_remoteproc_carveouts( tree, subnode, verbose)
    if ret == False:
        return ret


    return True
